      SUBROUTINE LESLIL(GETAV,M,N,APPLES,T0,TE,DT,Y0,
     *                      TOLQ,TOLL,IPAR,FWORK,IFLAG,INARR,REARR)

C
C ***** WARNING: before calling LESLIS for the very first time,
C                the user must call INIT, which will be called only
C                once, regardless of integration options
C                In INIT, the code checks for valid Input, sets
C                up workspace and defaults.  The user must call
C                INIT after having set the Input data, in particular
C                the array IPAR.
C                See the sample driver "DRIV" for how this is
C                done *****
C
C	LESLIL approximates N Lyapunov exponents of a 
C       linear system Y'=A(t)Y on the interval [T0,TE]
C
C	LESLIL refers to: Lyapunov ExponentS LInear Large
C       in that it approximates Lyapunov exponents of a 
C       linear system for which it is only required to have the
C       action of the coefficient matrix A(t) on a vector v: A(t)v
C       Presumably, this is the case for Large systems.
C       In case of small systems, for which the coefficient
C       matrix can be stored, and is known, then the user
C       should consider using LESLIS (Lyapunov Exponents
C       Linear Small).  For nonlinear systems, see the
C       documentation of LESNLL and LESNLS. 
C
C	LESLIL is an interface to subroutines which approximate
C       the Lyapunov exponents by using either the discrete or 
C       the continuous QR method with a number of integration 
C       options.  See IPAR, which is the array allowing the user 
C       to specify which options are desired.
C
C       In a nutshell, one can proceed in fixed or variable 
C       step size.  In the latter case, different options of
C       local error control are allowed.
C       No approximation of solution trajectory is attempted.
C
C       The details of the algorithms used can be found in 
C
C       [-] L. Dieci, E. Van Vleck
C           "LESLIS and LESLIL: Codes for approximating Lyapunov 
C            exponents of linear systems", Technical Report available 
C            from the web-page of either author below.
C
C	Authors:
C		Luca Dieci
C		School of Mathematics
C 		Georgia Institute of Technology
C		Atlanta, GA
C		dieci@math.gatech.edu
C               web: www.math.gatech.edu/~dieci
C
C		Erik S. Van Vleck
C		Department of Mathematics
C		University of Kansas
C		Lawrence, KS
C		evanvleck@math.ukansas.edu
C               web: www.math.ukans.edu/~evanvleck
C
C	Inquiries, Bugs, Malfunctioning, Fixes, HowTo:
C
C		Luca Dieci
C		School of Mathematics
C 		Georgia Institute of Technology
C		Atlanta, GA, 30332
C		dieci@math.gatech.edu
C
C   ***** User Interface *****
C
C	Variables:
C
C       GETAV - subroutine that returns the coefficient matrix A 
C           times a vector v.  It must be declared external and 
C           have parameter list of the form
C                        GETAV(M,T,V,VDOT,INARR,REARR)
C           where M is the dimension of the system,
C           T is the value of the independent variable, 
C           V is the vector to be multiplied by A and 
C           VDOT is the result of A(t)*v
C           INARR and REARR are integer and double precision arrays
C              that the user can use for communication between the calling
C              program and the routine GETAV.  Treat them as dummy
C              arguments if they are not needed.
C       
C       M - integer. Number of rows, dimension of the problem
C           Must have 1.LE.M
C	
C       N - integer. Number of Lyapunov exponents desired
C           Must have 1.LE.N.LE.M
C
C       APPLES - double precision vector of dimension N. 
C           These are the approximate Lyapunov exponents.  For random 
C           initial conditions, Y0, and large interval where we compute, 
C           the LEs are typically ordered from largest to smallest.
C           DO NOT CHANGE content between calls.
C
C	T0 - initial time, double precision
C            Initial time on INPUT, final time reached on OUTPUT
C            Thus, on OUTPUT should have T0=TE when IPAR(2)=0
C
C	TE - Ending time, double precision.  
C            Must be different from T0.  Can have TE<T0 or TE>T0
C            But cannot change direction of integration between calls
C
C	DT - Chosen stepsize in fixed stepsize mode (IPAR(1)=1), 
C            stores the current stepsize in variable stepsize mode. 
C
C	Y0 - double precision, dimension (M,N).  On INPUT,
C            these are ICs on the fundamental matrix (see IPAR(4)).
C            On OUTPUT, it contains the Q-factor of the FMS at the
C            point we reached (typically, TE)
C
C	TOLQ - double precision scalar.
C             This is local error tolerance on the sup-norm error 
C             for the columns of Q (see IPAR(8) and IPAR(9)).  
C             [In constant stepsize mode (IPAR(1)=1), treat it as dummy argument]
C             In variable stepsize, must have 1.E-2>=TOLQ>=1.E-13 
C             RECOMMENDED values in the range [1.E-10, 1.E-3] if use
C             the Dormand-Prince pair (IPAR(8)=0,2,4) and in the range
C             [1.E-7, 1.E-3] if use the RK38 rule (IPAR(8)=1,2,5)
C
C	TOLL - double precision vector of dimension N.
C             These are local error tolerance on the Lyapunov exponents 
C             (see IPAR(8) and IPAR(9)).  
C             [In constant stepsize mode (IPAR(1)=1), treat it as dummy argument]
C             In variable stepsize, must have 1.E-2>=TOLL(I)>=1.E-13 
C             RECOMMENDED values in the range [1.E-10, 1.E-3] if use
C             the Dormand-Prince pair (IPAR(8)=0,2,4) and in the range
C             [1.E-7, 1.E-3] if use the RK38 rule (IPAR(8)=1,2,5)
C             [Given the fact that it is harder to approximate the smallest
C             exponents, user may want to ask for tighter tolerances for
C             largest exponents (the first few ones)]
C
C	IPAR - integer array of dimension 13.  This serves to communicate
C              to the code several integration options.
C              Some entries can be defaulted (indicated by DEFAULT below).
C	       IPAR(1) - INPUT: set IPAR(1)=0 for variable stepsize,
C                               set IPAR(1)=1 for fixed stepsize (in DT)
C                        [if IPAR(1).NE.1, code sets DEFAULT IPAR(1)=0]
C	       IPAR(2) - INPUT: from T0 to TE --> 0, in one step mode --> 1
C                        [if IPAR(2).ne.1, code sets DEFAULT IPAR(2)=0]
C                        One-step-mode means that the code returns control to  
C                        the user after a successful step in direction of TE
C	       IPAR(3) - INPUT: must be set to 1 before caling INIT, and
C                        also for very first call to the code
C                        (in one step mode or otherwise, and in constant 
C                        or variable stepsize mode)
C              IPAR(4) - INPUT: yes/no ICs on Y.  
C                        If IPAR(4) = 0, the DEFAULT, the code sets 
C                           ICs given by Y = [I_n 0]
C                        If IPAR(4) = 1, you must specify ICs in Y0
C              IPAR(5) - INPUT: dimension of FWORK; must have
C                        IPAR(5).GE.11*M*N+8*N+63
C	       IPAR(6) -  
C	       IPAR(7) -  
C              IPAR(8) - INPUT: this specifies which method and RK formula 
C                        to use.  Projection and Hybrid are different 
C                        implementations of the continuous QR method, while 
C                        Disc QR refers to the discrete QR method 
C                        (see the Technical Report quoted above).
C                        DP5 refers to the Dormand-Prince embedded 
C                        pair of order 5
C                        RK38 refers to an embedded pair of order 4 
C                        (the RK 3/8 rule)
C                        IPAR(8)=0 (DEFAULT) is Projection/DP5
C                        IPAR(8)=1 is Hybrid/DP5
C                        IPAR(8)=2 is Projection/RK38
C                        IPAR(8)=3 is Hybrid/RK38
C                        IPAR(8)=4 is Disc QR/DP5
C                        IPAR(8)=5 is Disc QR/RK38
C	       IPAR(9) - INPUT: In case in which IPAR(8)=0,1,2,3, IPAR(9)
C                        specifies if LEs are to be found by integration 
C                        with DP5 or RK38 on the nu-variables, or by 
C                        the composite trapezoidal rule.
C                        IPAR(9)=0, the DEFAULT, the LEs are to be found
C                           using DP5 or RK38 on the nu-variables
C                        IPAR(9)=1, the LEs are found by the composite 
C                           trapezoidal rule.
C                 -->    IPAR(9) cannot be set to 1 if IPAR(8)=4,5.
C                 -->    In case IPAR(9)=1, and IPAR(1)=0, error control 
C                        can only be performed on the Q-solution
C                        Hence, in variable stepsizes, we can set IPAR(9)=1 
C                        only with IPAR(8)=0,1,2,3 and IPAR(10)=1
C                 -->    N.B.: When you choose IPAR(9)=1, then 
C                        the schemes used for the Q-integration are not the 
C                        complete projection or hybrid method, but the 
C                        simple projection/hybrid methods.  
C              IPAR(10) - INPUT: In variable stepsize mode, IPAR(1)=0, 
C                        then IPAR(10) specifies if you want to do error control 
C                        on the Lyapunov exponents or on the Q-factor, or both.
C                 -->    If you set IPAR(8)=0,1,2,3 then you can choose
C                           IPAR(10)=0 --> error control on LEs  
C                           IPAR(10)=1 --> error control on Q 
C                           IPAR(10)=10 --> error control on LEs and Q 
C                        It must be appreciated that IPAR(10)=10 will
C                        generally lead to more accurate approximations for
C                        the LEs, since it amounts to essentially an error
C                        control per-unit-step on the LEs, while IPAR(10)=0
C                        is essentially an error-per-step control, and it
C                        is generally a bit less accurate but less expensive.
C                 -->    If you used IPAR(8)=0,1,2,3 and IPAR(9)=0, you can 
C                           choose IPAR(10)=0,1,10.  But, if you chose 
C                           IPAR(8)=0,1,2,3 and IPAR(9)=1, then can only use
C                           IPAR(10)=1
C                        When IPAR(1)=0 and IPAR(9)=0, the code sets 
C                        the DEFAULT IPAR(10)=0
C                 -->    If you chose IPAR(8)=4,5 then can only use IPAR(10)=0
C	       IPAR(11) - OUTPUT: set to 0 if code reached TE,
C                                otherwise it is set to 1
C              IPAR(12) - OUTPUT: NREJ (number of rejections) 
C              IPAR(13) - OUTPUT: NST (number of successful steps taken)
C
C	FWORK - double precision array of dimension IPAR(5)
C               Do not touch between calls
C
C	IFLAG - OUTPUT flag
C               IFLAG=0 successful return; either one step was completed
C                       in one step mode, or solutions reached TE
C               IFLAG=1 unsuccessful return; wrong dimension of FWORK
C               IFLAG=2 unsuccessful return: it is first call, but IPAR(3).ne.1
C               IFLAG=3 unsuccessful return; wrong value for TOLQ or TOLL 
C               IFLAG=4 mildly unsuccessful return; TE=T0, nothing done
C               IFLAG=5 mildly unsuccessful return; abs-value stepsize
C                       smaller than 100*eps (about 2.2e-14).
C                       No further integration allowed.
C                       Most likely this is caused by having transition
C                       matrix not full rank to working precision, and
C                       having unsuccessfully tried to reduce the stepsize
C                       to the minimum stepsize.
C               IFLAG=6 unsuccessful return; wrong sign of DT in fixed 
C                       stepsize mode
C               IFLAG=7 unsuccessful return; wrong value of N
C               IFLAG=8 unsuccessful return; you gave ICs on Y0 not of full rank
C               IFLAG=9 unsuccessful return; wrong input: M.lt.1
C               IFLAG=10 unsuccessful return; something unexpected happened,
C                        most likely computed transition not full rank to
C                        working precision
C               IFLAG=11 
C               IFLAG=12 unsuccessful return.  You want to proceed in variable 
C                        step-size, with the cont-QR method, and want to 
C                        approximate the LEs by the composite trap-rule,  
C                        but still required error control on the LEs.  
C                        Check syntax of IPAR(9) and IPAR(10).
C               IFLAG=13 unsuccessful return.  Wrong request for variable stepsize
C                        discrete QR method.
C                        Either you want to approximate the LEs by the composite 
C                        trap-rule, or want to control the error on Q.
C                        Check syntax of IPAR(9) and IPAR(10).
C
C       INARR, REARR: integer and double precision arrays
C           that the user can use for communication between the calling
C           program and the routine GETAV.  Treat them as dummy
C           arguments if they are not needed.
C
C*****Author: L Dieci 
C*****Date: September 2, 2004
C
CCCCCCC DECLARE VARIABLES
      IMPLICIT NONE
CCCCCCC INPUT/OUTPUT
      INTEGER M,N
      DOUBLE PRECISION APPLES(*),T0,TE,DT,TOLQ,TOLL(*),REARR(*)
      INTEGER IPAR(*), INARR(*)
      DOUBLE PRECISION FWORK(*), Y0(M,N)
      INTEGER IFLAG
      EXTERNAL GETAV

CCCCCCC LOCAL
      DOUBLE PRECISION T,HOLD, QDNORM, TOLM
      INTEGER NREJ, INTDIR, IFIRST
      INTEGER I, J

CCCCCCC POINTERS, AND COMMON VARIABLES
      LOGICAL LAST
      DOUBLE PRECISION HMIN, HNEXT, TFIRST
      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      INTEGER IYPTR, IYHPTR, IQPTR, IBPTR
      INTEGER K1PTR, K2PTR, K3PTR, K4PTR, K5PTR, K6PTR
      INTEGER ISTAGE
      INTEGER INUPTR, INUHTP
      INTEGER K1NUPT, K2NUPT, K3NUPT, K4NUPT, K5NUPT, K6NUPT
      INTEGER IARKPT,ICRKPT,IBHTPT
      COMMON /STEP/ HMIN, HNEXT, TFIRST, INTDIR, IFIRST, LAST
      SAVE /STEP/
      COMMON /NSQUAX/ IYPTR, IQPTR, IYHPTR, IBPTR
      SAVE /NSQUAX/
      COMMON /RKPTRS/ K1PTR, K2PTR, K3PTR, K4PTR, K5PTR, K6PTR,
     *                ISTAGE
      SAVE /RKPTRS/
      COMMON /RNUPTR/INUPTR, INUHTP, K1NUPT, K2NUPT, K3NUPT,
     *              K4NUPT, K5NUPT, K6NUPT 
      SAVE /RNUPTR/
      COMMON /RKCF/ IARKPT, ICRKPT, IBHTPT
      SAVE /RKCF/
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/

      IFLAG = 0
      IF (IPAR(3).EQ.1) THEN
C FIRST TIME, INITIALIZE LES , AND SAVE Q FACTOR OF Y0 IN Y0 ITSELF
         IF (IPAR(4).EQ.1) THEN
            CALL MODGRS(M,N,Y0,M,FWORK(IYHPTR),IFLAG) 
            IF (IFLAG.NE.0) THEN
               IFLAG = 8
               RETURN
            END IF
           ELSE
            DO 5 J=1,N
               DO 2 I=1,M
                  Y0(I,J)=ZERO
 2             CONTINUE
               Y0(J,J)=ONE
 5          CONTINUE
         END IF
         DO 7 I=1,N
            APPLES(I)=ZERO
 7       CONTINUE
         IFIRST=1
         IPAR(3)=0
         IF (IPAR(1).EQ.0) THEN
CCCCCCC PREPARE INITIAL STEP SIZE IN VARIABLE STEP MODE
C IF IPAR(8)=0,2, WE USE Q-DOT, OTHERWISE WE USE Y-DOT
             IF (IPAR(8).EQ.0.OR.IPAR(8).EQ.2) THEN 
C QDOT RETURNS QDOT IN RK1
               CALL QDOT(GETAV,M,N,T0,FWORK(IBPTR),Y0,FWORK(K1PTR),
     1                INARR,REARR)
              ELSE
C YDOT RETURNS YDOT IN RK1
               CALL YDOT(GETAV,M,N,T0,Y0,FWORK(K1PTR),INARR,REARR)
            ENDIF
C HAVE Y-DOT OR Q-DOT IN FWORK(K1PTR).  FIND ITS MAX-NORM
            QDNORM=ZERO
            DO 10 J=1,N
               DO 10 I=1,M
 10               QDNORM=DMAX1(QDNORM,DABS(FWORK(K1PTR+(I-1)*N+J-1)))
C TAKE DT=(5-TH ROOT OF TOL) FOR RKDP OR
C      DT=(4-TH ROOT OF TOL) FOR RK38
C WEIGHTED AGAINST DERIVATIVE OF Q
            IF (IPAR(10).EQ.0) THEN
               TOLM=TOLL(1)
               DO 11 I=2,N
 11               TOLM=DMIN1(TOLM,TOLL(I))
              ELSE
               TOLM=TOLQ
            ENDIF
            IF (IPAR(8).EQ.0.OR.IPAR(8).EQ.1.OR.IPAR(8).EQ.4) THEN 
               DT=INTDIR*TOLM**EXP1/(DMAX1(ONE,QDNORM))
              ELSE
               DT=INTDIR*TOLM**EXP2/(DMAX1(ONE,QDNORM))
            END IF
            IF (INTDIR*(T0+DT).GT.INTDIR*TE) DT=TE-T0
            IF (DABS(DT).LE.HMIN) THEN
               IFLAG=5
               RETURN
            END IF
         END IF
      END IF
CCCCCCC LOOP
      T = T0
 20   CONTINUE  
      NREJ=0
C MONITOR IF THIS IS (PRESUMABLY) LAST STEP
C ADJUST STEPSIZE AS NEEDED, AND STORE 
C IN HNEXT THE STEP WHICH WOULD HAVE BEEN USED
      IF (LAST) THEN
         LAST=.FALSE.
         IF (IPAR(1).EQ.0) DT=HNEXT
      ENDIF
      IF (DABS(DT).GE.DABS(TE-T)) THEN
         LAST=.TRUE.
         HNEXT=DT
         IF (IPAR(1).EQ.0) DT=INTDIR*DABS(TE-T)
        ELSE
         LAST=.FALSE.
      ENDIF
      IF ((DABS(DT).LE.HMIN).AND.(.NOT.LAST)) THEN
         IFLAG=5
         RETURN
      ENDIF
C CALL TO ONE OF THE DIFFERENT METHODS.
      IF (IPAR(8).EQ.0) THEN 
CCCCCCC PQDPLL
         CALL PQDPLL(GETAV,M,N,T,DT,HOLD,Y0,APPLES,TOLQ,TOLL,
     *    FWORK(IQPTR),FWORK(IYHPTR),FWORK(IBPTR),
     *    FWORK(K1PTR),FWORK(K2PTR),
     *    FWORK(K3PTR),FWORK(K4PTR),FWORK(K5PTR), FWORK(K6PTR),
     *    FWORK(ISTAGE), 
     *    FWORK(INUPTR),FWORK(INUHTP),FWORK(K1NUPT),FWORK(K2NUPT), 
     *    FWORK(K3NUPT), FWORK(K4NUPT), FWORK(K5NUPT), 
     *    FWORK(K6NUPT),FWORK(IARKPT),FWORK(ICRKPT),
     *    FWORK(IBHTPT),IPAR,NREJ,IFLAG,INARR,REARR)
       ELSEIF (IPAR(8).EQ.1) THEN
CCCCCCC HQDPLL
         CALL HQDPLL(GETAV,M,N,T,DT,HOLD,Y0,APPLES,TOLQ,TOLL,
     *      FWORK(IYPTR),FWORK(IYHPTR),FWORK(IQPTR), FWORK(IBPTR),
     *      FWORK(K1PTR),FWORK(K2PTR),
     *      FWORK(K3PTR),FWORK(K4PTR),FWORK(K5PTR), FWORK(K6PTR),
     *      FWORK(ISTAGE), 
     *      FWORK(INUPTR),FWORK(INUHTP),FWORK(K1NUPT),FWORK(K2NUPT), 
     *      FWORK(K3NUPT), FWORK(K4NUPT), FWORK(K5NUPT),
     *       FWORK(K6NUPT),FWORK(IARKPT),FWORK(ICRKPT),
     *      FWORK(IBHTPT),IPAR,NREJ,IFLAG,INARR,REARR)
       ELSEIF (IPAR(8).EQ.2) THEN
CCCCCCC PQ38LL
         CALL PQ38LL(GETAV,M,N,T,DT,HOLD,Y0,APPLES,TOLQ,TOLL,
     *    FWORK(IQPTR),FWORK(IYHPTR), FWORK(IBPTR),
     *    FWORK(K1PTR),FWORK(K2PTR),
     *    FWORK(K3PTR),FWORK(K4PTR),FWORK(ISTAGE),
     *    FWORK(INUPTR),FWORK(INUHTP),FWORK(K1NUPT),FWORK(K2NUPT), 
     *    FWORK(K3NUPT), FWORK(K4NUPT),  
     *    FWORK(IARKPT),FWORK(ICRKPT),
     *    FWORK(IBHTPT),IPAR,NREJ,IFLAG,INARR,REARR)
       ELSEIF (IPAR(8).EQ.3) THEN
CCCCCCC HQ38LL
         CALL HQ38LL(GETAV,M,N,T,DT,HOLD,Y0,APPLES,TOLQ,TOLL,
     *    FWORK(IYPTR),FWORK(IYHPTR), FWORK(IQPTR), FWORK(IBPTR),
     *    FWORK(K1PTR),FWORK(K2PTR),
     *    FWORK(K3PTR),FWORK(K4PTR),FWORK(ISTAGE),
     *    FWORK(INUPTR),FWORK(INUHTP),FWORK(K1NUPT),FWORK(K2NUPT), 
     *    FWORK(K3NUPT), FWORK(K4NUPT),  
     *    FWORK(IARKPT),FWORK(ICRKPT),
     *    FWORK(IBHTPT),IPAR,NREJ,IFLAG,INARR,REARR)
       ELSEIF (IPAR(8).EQ.4) THEN
CCCCCCC YDPLL
         CALL YDPLL(GETAV,M,N,T,DT,HOLD,Y0,APPLES,TOLL,
     *    FWORK(IYPTR),FWORK(IYHPTR),FWORK(K1PTR),FWORK(K2PTR),
     *    FWORK(K3PTR),FWORK(K4PTR),FWORK(K5PTR), FWORK(K6PTR),
     *    FWORK(ISTAGE), FWORK(IARKPT),FWORK(ICRKPT),
     *    FWORK(IBHTPT),IPAR(1),NREJ,IFLAG,INARR,REARR)
       ELSEIF (IPAR(8).EQ.5) THEN
CCCCCCC Y38LL
         CALL Y38LL(GETAV,M,N,T,DT,HOLD,Y0,APPLES,TOLL,
     *    FWORK(IYPTR),FWORK(IYHPTR),FWORK(K1PTR),FWORK(K2PTR),
     *    FWORK(K3PTR),FWORK(K4PTR),FWORK(ISTAGE),
     *    FWORK(IARKPT),FWORK(ICRKPT),
     *    FWORK(IBHTPT),IPAR(1),NREJ,IFLAG,INARR,REARR)
      ENDIF
C CHECK IF ALL WENT WELL, UPDATE COUNTERS, RETURN CONTROL IF NEEDED
      IF (IFLAG.NE.0) RETURN  
      IPAR(12) = IPAR(12)+NREJ
      IPAR(13) = IPAR(13)+1
CCCCCCCCCC UPDATE FOR NEXT STEP
      T = T + HOLD
      T0=T
C CHECK IF REACHED DESIRED END POINT TE 
      IF (DABS(T).LT.DABS(TE)) THEN
         IPAR(11)=1
        ELSE
         IPAR(11)=0
      END IF
C CHECK IF FAILED ON WHAT HAD TO BE LAST STEP 
      IF((LAST).AND.(IPAR(11).EQ.1)) LAST=.FALSE.
CCCCCCC RETURN IF IN ONESTEP MODE 
      IF ((IPAR(2).EQ.0).AND.(IPAR(11).EQ.1)) GOTO 20 

      RETURN
      END
CCCCC END OF LESLIL CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: INIT
C
C*****Purpose
C
C checks for correctness of INPUT,
C sets up pointers, RK coefficients, and other quantities
C which need to be computed only once
C
C input: 
C     M, dimension of coefficient matrix
C     N, number of exponents to approximate
C     IPAR, integer vector of length 12 with information
C        on what to do
C     T0, TE, DT : initial and final times and stepsize
C        if wants to proceed in constant stepsize mode
C     TOLQ, TOLL: error tolerances on Q and the LEs
C     FWORK : work array with memory to be split
C 
C output:
C     indices and quantities to retain in common blocks
C     IFLAG = 0  all went well.  If IFLAG is not 0, then
C        some mishap occurred
C
C*****Authors: L Dieci & ES Van Vleck
C*****Date: May 20, 2004
C
CCCCCC INIT CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE INIT(M,N,IPAR,T0,TE,DT,TOLQ,TOLL,FWORK,IFLAG)
      IMPLICIT NONE
CCCCCCC INPUT/OUTPUT
      INTEGER M,N
      INTEGER IPAR(*)
      DOUBLE PRECISION T0, TE, DT, TOLQ, TOLL(*), FWORK(*)
      INTEGER IFLAG

      DOUBLE PRECISION TOLMIN, TOLMAX, ACC
      INTEGER I
      
CCCCCCC POINTERS, AND COMMON VARIABLES
      LOGICAL LAST
      DOUBLE PRECISION HMIN, HNEXT, TFIRST
      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      INTEGER INTDIR, IFIRST
      INTEGER IYPTR, IYHPTR, IQPTR, IBPTR
      INTEGER K1PTR, K2PTR, K3PTR, K4PTR, K5PTR, K6PTR
      INTEGER ISTAGE
      INTEGER INUPTR, INUHTP
      INTEGER K1NUPT, K2NUPT, K3NUPT, K4NUPT, K5NUPT, K6NUPT
      INTEGER IARKPT,ICRKPT,IBHTPT
      COMMON /STEP/ HMIN, HNEXT, TFIRST, INTDIR, IFIRST, LAST
      SAVE /STEP/
      COMMON /NSQUAX/ IYPTR, IQPTR, IYHPTR, IBPTR
      SAVE /NSQUAX/
      COMMON /RKPTRS/ K1PTR, K2PTR, K3PTR, K4PTR, K5PTR, K6PTR, 
     *                ISTAGE
      SAVE /RKPTRS/
      COMMON /RNUPTR/INUPTR, INUHTP, K1NUPT, K2NUPT, K3NUPT,
     *              K4NUPT, K5NUPT, K6NUPT 
      SAVE /RNUPTR/
      COMMON /RKCF/ IARKPT, ICRKPT, IBHTPT
      SAVE /RKCF/
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/

      IFLAG = 0
      LAST=.FALSE.
C CHECK CORRECTENESS OF INPUT ON M AND N
      IF (M.LT.1) THEN
         IFLAG=9
         RETURN
      END IF
      IF ((N.GT.M).OR.(N.LT.1)) THEN
         IFLAG=7
         RETURN
      END IF
CCCCCCC CHECK SIZE OF FWORK 
CC       CURRENTLY NEED DIM(FWORK).GE.11*M*N+8*N+63
      IF (IPAR(5).LT.(11*M*N+8*N+63)) THEN
         IFLAG=1
         RETURN
      END IF
C CHECK VALUES OF IPAR AND SET DEFAULTS IF NEEDED
      IF (IPAR(1).NE.1) IPAR(1)=0
      IF (IPAR(2).NE.1) IPAR(2)=0
      IF (IPAR(3).NE.1) THEN
         IFLAG=2
         RETURN
      END IF
      IF (IPAR(4).NE.1) IPAR(4)=0      
      IF ((IPAR(8).GT.0).AND.(IPAR(8).LT.6)) THEN
         IPAR(8) = IPAR(8)
        ELSE
         IPAR(8) = 0
      END IF
      IF (IPAR(9).EQ.1.AND.IPAR(8).GT.3) THEN
         IFLAG=13
         RETURN
      END IF
      IF (IPAR(9).NE.1) IPAR(9)=0      
      IF ((IPAR(10).NE.1).AND.(IPAR(10).NE.10)) IPAR(10)=0 
      IF ((IPAR(1).EQ.0).AND.(IPAR(8).GT.3.AND.IPAR(10).GT.0)) THEN
         IFLAG=13
         RETURN
      END IF
      IF ((IPAR(1).EQ.0).AND.(IPAR(9).EQ.1).AND.(IPAR(10).NE.1)) THEN
         IFLAG=12
         RETURN
      ENDIF
      IPAR(12)=0
      IPAR(13)=0
C SET UP VALUES OF CONSTANTS NEEDED IN COMMON BLOCK CONSTS
      ONE=1.0D0
      ZERO=0.0D0
C THE VALUES BELOW MAY BE CHANGED, BUT THIS IS DISCOURAGED
      EXP1=0.20D0
      EXP2=0.25D0
      HINCR=5.0D0
      HSFTY=0.8D0
      HDRSTC=0.2D0
      TOLMIN=1.0D-13
      TOLMAX=1.0D-2
C CHECK TOLQ, TOLL    
      IF (IPAR(1).EQ.0) THEN
         IF (IPAR(10).EQ.1.OR.IPAR(10).EQ.10) THEN
            IF ((TOLQ.LT.TOLMIN).OR.(TOLQ.GT.TOLMAX)) THEN
               IFLAG=3
               RETURN
            END IF
         ENDIF
         IF (IPAR(10).EQ.0.OR.IPAR(10).EQ.10) THEN
            DO 2 I=1,N
               IF (TOLL(I).LT.TOLMIN.OR.TOLL(I).GT.TOLMAX) THEN
                  IFLAG=3
                  RETURN
               END IF
 2          CONTINUE
         ENDIF
      END IF
C CHECK TO.NE.TE
C SET INTEGRATION DIRECTION AND SAVE VERY FIRST TIME
      IF (T0.EQ.TE) THEN
         IFLAG=4
         RETURN
        ELSE
         IF(TE.GT.T0)THEN
            INTDIR=1
           ELSE
            INTDIR=-1   
         END IF
         TFIRST=T0
      END IF
C CHECK INPUT IF FIXED STEPSIZE
      IF((IPAR(1).EQ.1) .AND. (INTDIR*DT.LT.ZERO)) THEN
         IFLAG=6
         RETURN
      ENDIF
C COMPUTE MINIMAL ALLOWED STEPSIZE AND EPS
      HMIN=ONE
 5    HMIN=0.5D0*HMIN
      ACC=ONE+HMIN
      IF (ACC.GT.ONE) GOTO 5
      TEPS=2.0D0*HMIN
      HMIN=200.0D0*HMIN
      IF ((IPAR(1).EQ.1).AND.(DABS(DT).LE.HMIN)) THEN
         IFLAG=5
         RETURN
      END IF
CCCCCCC GENERATE POINTERS TO BREAK UP FWORK 
C  COEFFICIENT MATRIX A AND TRANSITIONS Y AND YHAT IN /NSQUAX/  
      IYPTR = 1
      IQPTR=IYPTR+M*N
      IYHPTR = IQPTR+M*N
      IBPTR=IYHPTR+M*N
C  QUANTITIES NEEDED TO THE RKDP INTEGRATOR IN /RKPTRS/  
      K1PTR = IBPTR+M*N
      K2PTR = K1PTR+M*N
      K3PTR = K2PTR+M*N
      K4PTR = K3PTR+M*N
      K5PTR = K4PTR+M*N
      K6PTR = K5PTR+M*N
      ISTAGE = K6PTR+M*N
C  QUANTITIES NEEDED TO THE DP5 INTEGRATOR FOR NU IN /RNUPTR/
      INUPTR = ISTAGE+M*N
      INUHTP = INUPTR+N
      K1NUPT = INUHTP+N
      K2NUPT = K1NUPT+N
      K3NUPT = K2NUPT+N
      K4NUPT = K3NUPT+N
      K5NUPT = K4NUPT+N
      K6NUPT = K5NUPT+N
C COEFFICIENTS NEEDED TO THE RK INTEGRATOR IN /RKCF/  
      IARKPT = K6NUPT+N
      ICRKPT = IARKPT+49
      IBHTPT = ICRKPT+7
C COMPUTE RK COEFFICIENTS ONCE AND FOR ALL
      CALL CFRK(IPAR(8),FWORK(IARKPT),FWORK(ICRKPT),
     *           FWORK(IBHTPT))

      RETURN 
      END 
CCCCC END OF INIT CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: CFRK
C
C*****Purpose
C
C forms the coefficients for RKDP or RK38
C
C*****Author: L Dieci 
C*****Date: May 29, 2003
C
CCCCCCC CFRK CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE CFRK(IFORM,ARK,CRK,BHAT)

      IMPLICIT NONE

      INTEGER IFORM
      DOUBLE PRECISION ARK(7,7), CRK(7), BHAT(7)
      INTEGER I, J
      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/
       
C SET UP THE RKDP COEFFICIENTS

      DO 2 I=1,7
         BHAT(I)=ZERO
         CRK(I)=ZERO
         DO 2 J=1,7
            ARK(J,I)=ZERO
 2    CONTINUE
      IF ((IFORM.EQ.0).OR.(IFORM.EQ.1).OR.(IFORM.EQ.4)) THEN
         ARK(2,1)=0.2D0
         CRK(2)=ARK(2,1)
         ARK(3,1)=0.075D0
         ARK(3,2)=0.225D0
         CRK(3)=ARK(3,1)+ARK(3,2)
         ARK(4,1)=0.9777777777777777777D0
         ARK(4,2)=-3.733333333333333333D0
         ARK(4,3)=3.5555555555555555555D0
         CRK(4)=0.8D0
         ARK(5,1)=19372.D0/6561.D0
         ARK(5,2)=-25360.D0/2187.D0
         ARK(5,3)=64448.D0/6561.D0
         ARK(5,4)=-212.D0/729.D0
         CRK(5)=ARK(5,1)+ARK(5,2)+ARK(5,3)+ARK(5,4)
         ARK(6,1)=9017.D0/3168.D0
         ARK(6,2)=-355.D0/33.D0
         ARK(6,3)=46732.D0/5247.D0
         ARK(6,4)=49.D0/176.D0
         ARK(6,5)=-5103.D0/18656.D0
         CRK(6)=ARK(6,1)+ARK(6,2)+ARK(6,3)+ARK(6,4)+ARK(6,5)
         ARK(7,1)=0.0911458333333333D0
         ARK(7,3)=0.449236298292902D0
         ARK(7,4)=0.651041666666667D0
         ARK(7,5)=-0.322376179245283D0
         ARK(7,6)=0.130952380952381D0
         CRK(7)=ONE
         BHAT(1)=0.0899131944444444D0 
         BHAT(3)=0.4534890685834082D0 
         BHAT(4)=0.6140625D0 
         BHAT(5)=-0.2715123820754717D0 
         BHAT(6)=0.08904761904761904D0 
         BHAT(7)=0.025D0 
       ELSE 
         ARK(2,1)=ONE/3.D0
         CRK(2)=ARK(2,1)
         ARK(3,1)=-ONE/3.D0
         ARK(3,2)=ONE
         CRK(3)=ARK(3,1)+ARK(3,2)
         ARK(4,1)=ONE
         ARK(4,2)=-ONE
         ARK(4,3)=ONE
         CRK(4)=ONE
         ARK(5,1)=0.125D0
         ARK(5,2)=0.375D0
         ARK(5,3)=0.375D0
         ARK(5,4)=0.125D0
         CRK(5)=ONE
         BHAT(1)=0.08333333333333D0 
         BHAT(2)=0.5D0 
         BHAT(3)=0.25D0 
         BHAT(5)=0.16666666666667D0 
      ENDIF

      RETURN
      END
CCCCC END OF CFRK CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: YDOT
C
C*****Purpose
C
C forms the derivative of the Y-equation 
C (this is also used to select the first stepsize)
C
C input: 
C     GETAV, declared EXTERNAL, returns coefficient matrix times a vector
C     N, dimension of the system (coefficient matrix)
C     NQ, number of columns of Q (i.e., to be triangularized)
C     T0, initial value of t (where to get derivative)
C     YMAT, the matrix Q (i.e., the orthogonal factor of Y0)
C output:
C     YP, the first derivative of Y
C
C*****Authors: L Dieci & ES Van Vleck
C*****Date: May 29, 2003
C
CCCCCC YDOT CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE YDOT(GETAV,N,NQ,T0,YMAT,YP,INARR,REARR)
      IMPLICIT NONE

      EXTERNAL GETAV
      INTEGER N, NQ, INARR(*)
      DOUBLE PRECISION T0, YMAT(N,NQ), YP(N,NQ), REARR(*)

      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/

      INTEGER I, J

C FORM A*Q
      DO 7 J=1,NQ
         DO 2 I=1,N
 2          YP(I,J) = ZERO
         CALL GETAV(N,T0,YMAT(1,J),YP(1,J),INARR,REARR) 
 7    CONTINUE

      RETURN
      END
CCCCC END OF YDOT CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: QDOT
C
C*****Purpose
C
C forms the derivative of the Q-equation 
C
C input: 
C     GETAV, declared EXTERNAL, returns coefficient matrix times a vector
C     N, dimension of the coefficient matrix
C     NQ, number of columns of Q (i.e., to be triangularized)
C     T0, initial value of t (where to get derivative)
C     B, workspace
C     QMAT, the matrix Q
C output:
C     B, on diagonal it has diag(Q^TAQ)
C     QP, the first derivative of Q
C
C*****Authors: L Dieci & ES Van Vleck
C*****Date: May 29, 2003
C
CCCCCC QDOT CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE QDOT(GETAV,N,NQ,T0,B,QMAT,QP,INARR,REARR)
      IMPLICIT NONE

      EXTERNAL GETAV
      INTEGER N, NQ, INARR(*)
      DOUBLE PRECISION T0, QMAT(N,NQ), QP(N,NQ), B(N,NQ), REARR(*)

      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/

      INTEGER I, J

C FORM A*Q
      DO 7 J=1,NQ
         DO 2 I=1,N
 2          QP(I,J) = ZERO
         CALL GETAV(N,T0,QMAT(1,J),QP(1,J),INARR,REARR) 
 7    CONTINUE
C FORM Q^T*A*Q-S (S IS SKEW MATRIX)
      CALL DGEMM ('T', 'N', NQ, NQ, N, ONE, QMAT, N, QP, N,
     $                   ZERO, B, N )
C FINISH OFF Q^T*A*Q-S
      DO 11 I=1,NQ
         DO 11 J=1,NQ
            IF (I.LT.J) THEN 
               B(I,J)=B(I,J)+B(J,I)
              ELSE IF (I.GT.J) THEN 
               B(I,J)=0.0D0
            END IF
 11   CONTINUE
C UPDATE QDOT
      CALL DGEMM ('N', 'N', N, NQ, NQ, -1.0D0*ONE, QMAT, N, B, N,
     $                   ONE, QP, N )

      RETURN
      END
CCCCC END OF QDOT CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: PQDPLL
C
C*****Purpose
C
C performs the one step integration of Q'=F(A,Q), Q0 given
C and monitors errors on the one step approximations to the LEs
C updates stepsize and approximation of the LEs
C 
C       Method used: Projection continuous QR method with DP5 as 
C          integrator for Q, plus complete projection for Q (i.e., also
C          for stage values). As specified by
C          IPAR(9), two options are provided for 
C          approximating the LEs:
C          (i) a quadrature rule for the nu-variables of the 
C              same order as DP5 (the default), or
C          (ii) the composite trap-rule.  
C  
C          Code can proceed in fixed or variable
C          step size.  In the latter case, error control is 
C          performed on Q or on the Lyapunov exponents,
C          depending on IPAR(9).
C          No approximation of solution trajectory is attempted.
C
C*****Author: L Dieci 
C*****Date: May 29, 2004
C
CCCCCC PQDPLL CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE PQDPLL(GETAV,M,N,T,H,HOLD,Y0,RLYAP,TOLQ,TOLL,
     *          Q,QHAT,BMAT,RK1,RK2,RK3,RK4,RK5,RK6,STAGE, 
     *          RNU,RNUHAT,RKNU1,RKNU2,RKNU3,RKNU4,RKNU5,RKNU6,
     *          ARK,CRK,BHAT,IPAR,NREJ,IFLAG,INARR,REARR)

      IMPLICIT NONE

      EXTERNAL GETAV
      INTEGER M,N,IPAR(*),NREJ,IFLAG, INARR(*)
      DOUBLE PRECISION T,H,HOLD, REARR(*)
      DOUBLE PRECISION Y0(M,N), RLYAP(N), TOLQ, TOLL(*)
      DOUBLE PRECISION Q(M,N), QHAT(M,N)
      DOUBLE PRECISION BMAT(M,N)
      DOUBLE PRECISION RK1(M,N),RK2(M,N)
      DOUBLE PRECISION RK3(M,N),RK4(M,N),RK5(M,N)
      DOUBLE PRECISION RK6(M,N), STAGE(M,N)
      DOUBLE PRECISION RNU(N), RNUHAT(N)       
      DOUBLE PRECISION RKNU1(N), RKNU2(N), RKNU3(N), RKNU4(N)
      DOUBLE PRECISION RKNU5(N), RKNU6(N)
      DOUBLE PRECISION ARK(7,7), CRK(7), BHAT(7)

      LOGICAL PREVIO, LAST
      INTEGER INTDIR, IFIRST, IFXDPF, IHOWLE, IERRC, I, J
      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      DOUBLE PRECISION HNEW, HMIN, HNEXT, TFIRST
      DOUBLE PRECISION SCALE, RTOL, ERR
      COMMON /STEP/ HMIN, HNEXT, TFIRST, INTDIR, IFIRST, LAST
      SAVE /STEP/
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/

      IFXDPF=IPAR(1)
      IHOWLE=IPAR(9)
      IERRC=IPAR(10)
      PREVIO=.FALSE.
C PREVIO IS .FALSE. IF PREVIOUS STEP WAS SUCCESSFUL
C IT IS .TRUE. IF PREVIOUS STEP WAS A FAILURE
 2    CONTINUE
CCCCCCCCCC INTEGRATE FOR Y: RETURNS Q, QHAT,  NU AND NUHAT
      CALL PQRKDP(GETAV,PREVIO,M,N,IFIRST,IFXDPF,IHOWLE,IERRC,
     *        T,H,Y0,Q,QHAT,BMAT,RK1,RK2,RK3,RK4,RK5,RK6, 
     *        STAGE, RNU,RNUHAT,RKNU1,RKNU2,RKNU3,RKNU4,
     *        RKNU5,RKNU6,ARK,CRK,BHAT,IFLAG,INARR,REARR)
      IF (IFIRST.EQ.1) IFIRST=0
      IF (IFLAG.NE.0) THEN
         IF (IFXDPF.EQ.0) THEN
C TRY DRASTIC WAY TO RECOVER
            IF (DABS(H).GT.HMIN) THEN
               IFLAG=0
               H=INTDIR*HMIN
               PREVIO=.TRUE.
               NREJ=NREJ+1
               GOTO 2
              ELSE
               IFLAG=5
               RETURN
            ENDIF
           ELSE
            IFLAG=10
            RETURN
         END IF
      END IF
      IF (IFXDPF.EQ.0) THEN
C HERE WE PREDICT NEXT STEPSIZE HNEW, NEVER ALLOWING IT TO
C BE GREATER THAN FIVE TIMES CURRENT H 
C (THIS VALUE CAN BE CHANGED IN HINCR IN "INIT")
         HNEW=HINCR*H
         RTOL=ZERO
         IF (IERRC.EQ.0) THEN
            CALL NERRNU(N,RNU,RNUHAT,ERR,TOLL,ONE,ZERO)
           ELSE IF (IERRC.EQ.1) THEN
            CALL NERRQ(M,N,Q,QHAT,Y0,ERR,TOLQ,ONE,ZERO)
           ELSE 
            CALL NERRNU(N,RNU,RNUHAT,ERR,TOLL,ONE,ZERO)
            RTOL=DMAX1(RTOL,ERR)
            CALL NERRQ(M,N,Q,QHAT,Y0,ERR,TOLQ,ONE,ZERO)
            ERR=DMAX1(ERR,RTOL)
         END IF
C PREDICT NEW STEPSIZE
         IF (ERR.GT.ZERO)         
     1      HNEW=DMIN1(DABS(HNEW),DABS(HSFTY*H*(ONE/ERR)**EXP1))
C GO BACK IF FAILED UNLESS H TOO SMALL
         IF (ERR.GT.ONE) THEN
C AVOID DRASTIC REDUCTION OF STEPSIZE (NEVER LESS THAN 1/5)
C (THIS VALUE CAN BE CHANGED IN HDRSTC IN "INIT")
            H=INTDIR*DMAX1(DABS(HNEW),HDRSTC*DABS(H))
            NREJ=NREJ+1
            IF (INTDIR*H.LE.HMIN) THEN
               PRINT*,' H = ', H
               IFLAG = 5
               RETURN
            ENDIF
            PREVIO=.TRUE.
            GOTO 2
         ENDIF            
      END IF
C HAVE COMPLETED THE STEP WITH STEPSIZE H
C SAVE LAST SUCCESSFUL STEPSIZE
      HOLD=H
      IF (IFXDPF.EQ.0) THEN
C         HNEW=INTDIR*DMAX1(DABS(HNEW),DABS(H))
         IF((PREVIO).AND.(DABS(HNEW).GT.DABS(H)))HNEW=H
         H=HNEW
      END IF
C FINALLY, UPDATE THE APPROXIMATION FOR THE LES AND Q
      SCALE=ONE/(T+HOLD-TFIRST)      
      DO 25 I=1,N
         RLYAP(I)=((T-TFIRST)*SCALE)*RLYAP(I)+SCALE*RNU(I)
C LOWER ORDER APPROXIMATION
C         RLYAP(I)=((T-TFIRST)*SCALE)*RLYAP(I)+SCALE*RNUHAT(I)
 25   CONTINUE         
      DO 28 J=1,N
         DO 28 I=1,M
            Y0(I,J)=Q(I,J)
 28   CONTINUE

      RETURN
      END
CCCCC END OF PQDPLL CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: PQRKDP
C
C*****Purpose
C
C integrates on one step the solution of Q'=F(A(t),Q), Q0=Q0
C and returns one approximation, Q to the Q-factor
C in the QR factorization of the fundamental solution Y
C It also returns two approximations to the updates for
C the LEs, rnu and rnuhat, which can be used for error purposes
C Integrator is the Dormand-Prince pair of order 5 (4)
C Scheme is total projection method if ihowle=0, or 
C the projection of last value only if ihowle=1
C 
C*****Author: L Dieci 
C*****Date: May 17, 2004
C
CCCCCC PQRKDP CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE PQRKDP(GETAV,PREVIO,M,N,IFIRST,IFXDPF,IHOWLE,IERRC,
     *        T,H,Y0,Q,QHAT,BMAT,RK1,RK2,RK3,RK4,RK5,RK6,STAGE,
     *        RNU,RNUHAT,RKNU1,RKNU2,RKNU3,RKNU4,RKNU5,RKNU6,
     *        ARK,CRK,BHAT,IFLAG,INARR,REARR)

      IMPLICIT NONE

      EXTERNAL GETAV
      LOGICAL PREVIO
      INTEGER M, N, IFIRST, IFXDPF, IHOWLE, IERRC, IFLAG, INARR(*)
      DOUBLE PRECISION H, T, Y0(M,N), REARR(*)
      DOUBLE PRECISION Q(M,N),QHAT(M,N)
      DOUBLE PRECISION BMAT(M,N)
      DOUBLE PRECISION RK1(M,N),RK2(M,N)
      DOUBLE PRECISION RK3(M,N),RK4(M,N),RK5(M,N)
      DOUBLE PRECISION RK6(M,N), STAGE(M,N)
      DOUBLE PRECISION RNU(N),RNUHAT(N)
      DOUBLE PRECISION RKNU1(N),RKNU2(N),RKNU3(N)
      DOUBLE PRECISION RKNU4(N),RKNU5(N),RKNU6(N)
      DOUBLE PRECISION ARK(7,7), CRK(7), BHAT(7)
      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      INTEGER I, J
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/

C INTIALIZE Q TO OLD Q
C THE RNU'S ARE INITIALLY AT 0, SINCE WE TAKE ONE STEP OF INTEGRAL
      DO 2 J=1,N
         RNU(J)=ZERO
         RNUHAT(J)=ZERO
         DO 2 I=1,M
            Q(I,J)=Y0(I,J)
            QHAT(I,J)=Y0(I,J)
 2    CONTINUE
C FORM RK1, RK2, RK3, RK4, RK5, RK6, RK7 
C FORM ALSO THE STAGE VALUES AT H/5, 3H/10, 4H/5, 8H/9 AND H (NB THE LATTER IS
C NOT THE SOLUTION VALUE, BUT JUST THE STAGE VALUE)
C ... RK1 AND U15 ...
C VERY FIRST CALL WITH FIXED STEPSIZE, OR IF NOT FAILED STEP NEED RK1
C OTHERWISE, HAVE RK1 IN RK1 
CC **** SEE IF CAN ECONOMIZE HERE ****
      CALL QDOT(GETAV,M,N,T,BMAT,Q,RK1,INARR,REARR)
      DO 5 J=1,N
         DO 5 I=1,M
            STAGE(I,J) = Q(I,J) + H*ARK(2,1)*RK1(I,J)
 5    CONTINUE
      DO 10 I=1,N
 10       RKNU1(I)=BMAT(I,I)
      IF (IHOWLE.EQ.0) THEN
         CALL MODGRS(M,N,STAGE,M,BMAT,IFLAG) 
         IF (IFLAG.NE.0) THEN
            IFLAG=10
            RETURN
         ENDIF
      ENDIF
C ... RK2 AND U03 ...
      CALL QDOT(GETAV,M,N,T+CRK(2)*H,BMAT,STAGE,RK2,INARR,REARR)
      DO 12 J=1,N
         DO 12 I=1,M
            STAGE(I,J) = Q(I,J) + H*(ARK(3,1)*RK1(I,J)+
     1                 ARK(3,2)*RK2(I,J))
 12   CONTINUE
      IF (IHOWLE.EQ.0) THEN
         DO 20 I=1,N
 20         RKNU2(I)=BMAT(I,I)
         CALL MODGRS(M,N,STAGE,M,BMAT,IFLAG) 
         IF (IFLAG.NE.0) THEN
            IFLAG=10
            RETURN
         ENDIF
      ENDIF
C ... RK3 AND U45 ...
      CALL QDOT(GETAV,M,N,T+CRK(3)*H,BMAT,STAGE,RK3,INARR,REARR)
      DO 22 J=1,N
         DO 22 I=1,M
            STAGE(I,J) = Q(I,J) + H*(ARK(4,1)*RK1(I,J)+
     1                ARK(4,2)*RK2(I,J)+ARK(4,3)*RK3(I,J))
 22   CONTINUE
      IF (IHOWLE.EQ.0) THEN
         DO 23 I=1,N
 23         RKNU3(I)=BMAT(I,I)
         CALL MODGRS(M,N,STAGE,M,BMAT,IFLAG) 
         IF (IFLAG.NE.0) THEN
            IFLAG=10
            RETURN
         ENDIF
      ENDIF
C ... RK4 AND U89 
      CALL QDOT(GETAV,M,N,T+CRK(4)*H,BMAT,STAGE,RK4,INARR,REARR)
      DO 34 J=1,N
         DO 34 I=1,M
            STAGE(I,J) = Q(I,J) + H*(ARK(5,1)*RK1(I,J)+
     1                 ARK(5,2)*RK2(I,J)+
     2                 ARK(5,3)*RK3(I,J)+ARK(5,4)*RK4(I,J))
 34   CONTINUE
      IF (IHOWLE.EQ.0) THEN
         DO 40 I=1,N
 40         RKNU4(I)=BMAT(I,I)
         CALL MODGRS(M,N,STAGE,M,BMAT,IFLAG) 
         IF (IFLAG.NE.0) THEN
            IFLAG=10
            RETURN
         ENDIF
      ENDIF
C ... RK5 AND U1 (STAGE VALUE AT H) ...
      CALL QDOT(GETAV,M,N,T+CRK(5)*H,BMAT,STAGE,RK5,INARR,REARR)
      DO 47 J=1,N
         DO 47 I=1,M
            STAGE(I,J) = Q(I,J) + H*(ARK(6,1)*RK1(I,J)+
     1                ARK(6,2)*RK2(I,J)+ARK(6,3)*RK3(I,J)+
     2                ARK(6,4)*RK4(I,J)+ARK(6,5)*RK5(I,J))
 47   CONTINUE
      IF (IHOWLE.EQ.0) THEN
         DO 50 I=1,N
 50         RKNU5(I)=BMAT(I,I)
         CALL MODGRS(M,N,STAGE,M,BMAT,IFLAG) 
         IF (IFLAG.NE.0) THEN
            IFLAG=10
            RETURN
         ENDIF
      ENDIF
C ... RK6, SOLUTION AT H ...
      CALL QDOT(GETAV,M,N,T+H,BMAT,STAGE,RK6,INARR,REARR)
      IF (IHOWLE.EQ.0) THEN
         DO 55 I=1,N
 55         RKNU6(I)=BMAT(I,I)
      ENDIF
      DO 60 J=1,N
         IF (IHOWLE.EQ.0) THEN 
            RNU(J)=RNU(J)+H*(ARK(7,1)*RKNU1(J)+
     1         ARK(7,3)*RKNU3(J)+ARK(7,4)*RKNU4(J)+
     2         ARK(7,5)*RKNU5(J)+ARK(7,6)*RKNU6(J))
         ENDIF
         DO 60 I=1,M
            Q(I,J) = Q(I,J) + H*(ARK(7,1)*RK1(I,J)+
     1         ARK(7,3)*RK3(I,J)+ARK(7,4)*RK4(I,J)+
     2         ARK(7,5)*RK5(I,J)+ARK(7,6)*RK6(I,J))
 60   CONTINUE
C ... RK7 (STORED IN RK2), AND COMPARISON SOLUTION QHAT AND NUHAT
C TO BE USED FOR ERROR MONITOR/STEP-SIZE SELECTION
      IF (IHOWLE.EQ.0) THEN
         CALL MODGRS(M,N,Q,M,BMAT,IFLAG) 
         IF (IFLAG.NE.0) THEN
            IFLAG=10
            RETURN
         ENDIF
      ENDIF
      IF (((IERRC.EQ.1).OR.(IERRC.EQ.10)).AND.IFXDPF.EQ.0) THEN
         CALL QDOT(GETAV,M,N,T+H,BMAT,Q,RK2,INARR,REARR)
         IF (IHOWLE.EQ.1) THEN
            CALL MODGRS(M,N,Q,M,STAGE,IFLAG) 
            IF (IFLAG.NE.0) THEN
               IFLAG=10
               RETURN
            ENDIF
            CALL YDOT(GETAV,M,N,T+H,Q,STAGE,INARR,REARR)
            CALL DGQTB(M,M,N,Q,STAGE,M,BMAT,M)
         ENDIF
        ELSE
         IF (IHOWLE.EQ.1) THEN
            CALL MODGRS(M,N,Q,M,BMAT,IFLAG) 
            IF (IFLAG.NE.0) THEN
               IFLAG=10
               RETURN
            ENDIF
         ENDIF
         CALL YDOT(GETAV,M,N,T+H,Q,STAGE,INARR,REARR)
         CALL DGQTB(M,M,N,Q,STAGE,M,BMAT,M)
      ENDIF
      DO 61 I=1,N
 61      RKNU2(I)=BMAT(I,I)
      DO 70 J=1,N 
         IF (IHOWLE.EQ.0.AND.IFXDPF.EQ.0.AND.IERRC.NE.1) THEN
            RNUHAT(J)=RNUHAT(J)+H*(BHAT(1)*RKNU1(J)+
     1       BHAT(3)*RKNU3(J)+BHAT(4)*RKNU4(J)+
     2       BHAT(5)*RKNU5(J)+BHAT(6)*RKNU6(J)+BHAT(7)*RKNU2(J))
            ELSEIF (IHOWLE.EQ.1) THEN
             CALL TRAPRL(RKNU1(J),RKNU2(J),H,RNU(J))
         ENDIF
         IF (((IERRC.EQ.1).OR.(IERRC.EQ.10)).AND.IFXDPF.EQ.0) THEN
            DO 65 I=1,M
               QHAT(I,J)=QHAT(I,J)+H*(BHAT(1)*RK1(I,J)+
     1           BHAT(3)*RK3(I,J)+BHAT(4)*RK4(I,J)+
     2           BHAT(5)*RK5(I,J)+BHAT(6)*RK6(I,J)+BHAT(7)*RK2(I,J))
 65         CONTINUE
         ENDIF
 70   CONTINUE
      IF (((IERRC.EQ.1).OR.(IERRC.EQ.10)).AND.IFXDPF.EQ.0) THEN
         CALL MODGRS(M,N,QHAT,M,BMAT,IFLAG) 
         IF (IFLAG.NE.0) THEN
            IFLAG=10
            RETURN
         ENDIF
      ENDIF

      RETURN
      END
CCCCC END OF PQRKDP CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: HQDPLL
C
C*****Purpose
C
C performs the one step integration of Y'=A(t)Y, Y0=Q0
C and monitors errors on the one step approximations to the LEs
C updates stepsize and approximation of the LEs
C 
C       Method used: Hybrid continuous QR method with DP5 as 
C          integrator for Y, and extraction of Q.  As specified by
C          IPAR(9), two options are provided for 
C          approximating the LEs:
C          (i) a quadrature rule for the nu-variables of the 
C              same order as DP5 (the default), or
C          (ii) the composite trap-rule.  
C
C          Code can proceed in fixed or variable step size.  
C          In the latter case, error control is 
C          performed on the Lyapunov exponents only or on the Q-factor,
C          depending on IPAR(9),
C          No approximation of solution trajectory is attempted.
C
C*****Author: L Dieci 
C*****Date: May 29, 2004
C
CCCCCC HQDPLL CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE HQDPLL(GETAV,M,N,T,H,HOLD,Y0,RLYAP,TOLQ,TOLL,
     *          Y,YHAT,QMAT,BMAT,RK1,RK2,RK3,RK4,RK5,RK6,STAGE, 
     *          RNU,RNUHAT,RKNU1,RKNU2,RKNU3,RKNU4,RKNU5,RKNU6,
     *          ARK,CRK,BHAT,IPAR,NREJ,IFLAG,INARR,REARR)

      EXTERNAL GETAV
      INTEGER M,N,IPAR(*), NREJ,IFLAG,INARR(*)
      DOUBLE PRECISION T,H,HOLD,REARR(*)
      DOUBLE PRECISION Y0(M,N), RLYAP(N), TOLQ, TOLL(*)
      DOUBLE PRECISION Y(M,N), YHAT(M,N)
      DOUBLE PRECISION QMAT(M,N), BMAT(M,N)
      DOUBLE PRECISION RK1(M,N),RK2(M,N)
      DOUBLE PRECISION RK3(M,N),RK4(M,N),RK5(M,N)
      DOUBLE PRECISION RK6(M,N), STAGE(M,N)
      DOUBLE PRECISION RNU(N), RNUHAT(N)       
      DOUBLE PRECISION RKNU1(N), RKNU2(N), RKNU3(N), RKNU4(N)
      DOUBLE PRECISION RKNU5(N), RKNU6(N)
      DOUBLE PRECISION ARK(7,7), CRK(7), BHAT(7)

      LOGICAL PREVIO, LAST
      INTEGER INTDIR, IFIRST, IFXDPF, IHOWLE, IERRC, I, J
      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      DOUBLE PRECISION HNEW, HMIN, HNEXT, TFIRST
      DOUBLE PRECISION SCALE, RTOL, ERR
      COMMON /STEP/ HMIN, HNEXT, TFIRST, INTDIR, IFIRST, LAST
      SAVE /STEP/
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/

      IFXDPF=IPAR(1)
      IHOWLE=IPAR(9)
      IERRC=IPAR(10)
      PREVIO=.FALSE.
C PREVIO IS .FALSE. IF PREVIOUS STEP WAS SUCCESSFUL
C IT IS .TRUE. IF PREVIOUS STEP WAS A FAILURE
 2    CONTINUE
CCCCCCCCCC INTEGRATE FOR Y: RETURNS Y AND YHAT
      CALL HQRKDP(GETAV,PREVIO,M,N,IFIRST,IFXDPF,IHOWLE,IERRC,
     *        T,H,Y0,Y,YHAT,QMAT,BMAT,RK1,RK2,RK3,RK4,RK5,RK6, 
     *        STAGE, RNU,RNUHAT,RKNU1,RKNU2,RKNU3,RKNU4,
     *        RKNU5,RKNU6,ARK,CRK,BHAT,IFLAG,INARR,REARR)
      IF (IFIRST.EQ.1) IFIRST=0
      IF (IFLAG.NE.0) THEN
         IF (IFXDPF.EQ.0) THEN
C TRY DRASTIC WAY TO RECOVER
            IF (DABS(H).GT.HMIN) THEN
               IFLAG=0
               H=INTDIR*HMIN
               PREVIO=.TRUE.
               NREJ=NREJ+1
               GOTO 2
              ELSE
               IFLAG=5
               RETURN
            ENDIF
           ELSE
            IFLAG=10
            RETURN
         END IF
      END IF
      IF (IFXDPF.EQ.0) THEN
C HERE WE PREDICT NEXT STEPSIZE HNEW, NEVER ALLOWING IT TO
C BE GREATER THAN FIVE TIMES CURRENT H 
C (THIS VALUE CAN BE CHANGED IN HINCR IN "INIT")
         HNEW=HINCR*H
         RTOL=ZERO
         IF (IERRC.EQ.0) THEN
            CALL NERRNU(N,RNU,RNUHAT,ERR,TOLL,ONE,ZERO)
           ELSE IF (IERRC.EQ.1) THEN
            CALL NERRQ(M,N,Y,YHAT,Y0,ERR,TOLQ,ONE,ZERO)
           ELSE 
            CALL NERRNU(N,RNU,RNUHAT,ERR,TOLL,ONE,ZERO)
            RTOL=DMAX1(RTOL,ERR)
            CALL NERRQ(M,N,Y,YHAT,Y0,ERR,TOLQ,ONE,ZERO)
            ERR=DMAX1(ERR,RTOL)
         END IF
C PREDICT NEW STEPSIZE
         IF (ERR.GT.ZERO)         
     1      HNEW=DMIN1(DABS(HNEW),DABS(HSFTY*H*(ONE/ERR)**EXP1))
C GO BACK IF FAILED UNLESS H TOO SMALL
         IF (ERR.GT.ONE) THEN
C AVOID DRASTIC REDUCTION OF STEPSIZE (NEVER LESS THAN 1/5)
C (THIS VALUE CAN BE CHANGED IN HDRSTC IN "INIT")
            H=INTDIR*DMAX1(DABS(HNEW),HDRSTC*DABS(H))
            NREJ=NREJ+1
            IF (INTDIR*H.LE.HMIN) THEN
               PRINT*,' H = ', H
               IFLAG = 5
               RETURN
            ENDIF
            PREVIO=.TRUE.
            GOTO 2
         ENDIF            
      END IF
C HAVE COMPLETED THE STEP WITH STEPSIZE H
C SAVE LAST SUCCESSFUL STEPSIZE
      HOLD=H
      IF (IFXDPF.EQ.0) THEN
C         HNEW=INTDIR*DMAX1(DABS(HNEW),DABS(H))
         IF((PREVIO).AND.(DABS(HNEW).GT.DABS(H)))HNEW=H
         H=HNEW
      END IF
C FINALLY, UPDATE THE APPROXIMATION FOR THE LES AND Q
      SCALE=ONE/(T+HOLD-TFIRST)      
      DO 25 I=1,N
         RLYAP(I)=((T-TFIRST)*SCALE)*RLYAP(I)+SCALE*RNU(I)
C LOWER ORDER APPROXIMATION
C         RLYAP(I)=((T-TFIRST)*SCALE)*RLYAP(I)+SCALE*RNUHAT(I)
 25   CONTINUE         
      DO 28 J=1,N
         DO 28 I=1,M
            Y0(I,J)=Y(I,J)
 28   CONTINUE

      RETURN
      END
CCCCC END OF HQDPLL CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: HQRKDP
C
C*****Purpose
C
C integrates on one step the solution of Y'=A(t)Y, Y0=Q0
C and returns two approximations, Y and YHAT to the Q-factor
C in the QR factorization of Y
C which can be used for error purposes
C Integrator is the Dormand-Prince pair of order 5 (4)
C 
C*****Author: L Dieci 
C*****Date: May 29, 2004
C
CCCCCC HQRKDP CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE HQRKDP(GETAV,PREVIO,M,N,IFIRST,IFXDPF,IHOWLE,IERRC,
     *        T,H,Y0,Y,YHAT,QMAT,BMAT,RK1,RK2,RK3,RK4,RK5,RK6,STAGE,
     *        RNU,RNUHAT,RKNU1,RKNU2,RKNU3,RKNU4,RKNU5,RKNU6,
     *        ARK,CRK,BHAT,IFLAG,INARR,REARR)

      IMPLICIT NONE

      EXTERNAL GETAV
      LOGICAL PREVIO
      INTEGER M, N, IFIRST, IFXDPF, IHOWLE, IERRC, IFLAG, INARR(*)
      DOUBLE PRECISION H, T, Y0(M,N), REARR(*)
      DOUBLE PRECISION Y(M,N),YHAT(M,N)
      DOUBLE PRECISION QMAT(M,N), BMAT(M,N)
      DOUBLE PRECISION RK1(M,N),RK2(M,N)
      DOUBLE PRECISION RK3(M,N),RK4(M,N),RK5(M,N)
      DOUBLE PRECISION RK6(M,N), STAGE(M,N)
      DOUBLE PRECISION RNU(N),RNUHAT(N)
      DOUBLE PRECISION RKNU1(N),RKNU2(N),RKNU3(N)
      DOUBLE PRECISION RKNU4(N),RKNU5(N),RKNU6(N)
      DOUBLE PRECISION ARK(7,7), CRK(7), BHAT(7)
      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      INTEGER I, J
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/

C INTIALIZE Y TO OLD Q
C THE RNU'S ARE INITIALLY AT 0, SINCE WE TAKE ONE STEP OF INTEGRAL
      DO 2 J=1,N
         RNU(J)=ZERO
         RNUHAT(J)=ZERO
         DO 2 I=1,M
            Y(I,J)=Y0(I,J)
            YHAT(I,J)=Y0(I,J)
 2    CONTINUE
C FORM RK1, RK2, RK3, RK4, RK5, RK6, RK7 
C FORM ALSO THE STAGE VALUES AT H/5, 3H/10, 4H/5, 8H/9 AND H (NB THE LATTER IS
C NOT THE SOLUTION VALUE, BUT JUST THE STAGE VALUE)
C ... RK1 AND U15 ...
C VERY FIRST CALL WITH FIXED STEPSIZE, OR IF NOT FAILED STEP NEED RK1
C OTHERWISE, HAVE RK1 IN RK1 
      IF ((IFIRST.EQ.1.AND.IFXDPF.EQ.1).OR.(.NOT.PREVIO)) THEN
         CALL YDOT(GETAV,M,N,T,Y0,RK1,INARR,REARR)
      END IF
      CALL DGQTB(M,M,N,Y,RK1,M,BMAT,M)
      DO 4 I=1,N
 4       RKNU1(I)=BMAT(I,I)
      DO 10 J=1,N
         DO 10 I=1,M
            STAGE(I,J) = Y(I,J) + H*ARK(2,1)*RK1(I,J)
 10   CONTINUE
C ... RK2 AND U03 ...
      CALL YDOT(GETAV,M,N,T+CRK(2)*H,STAGE,RK2,INARR,REARR)
      IF (IHOWLE.EQ.1) GOTO 14
      DO 11 J=1,N
         DO 11 I=1,M
 11         QMAT(I,J)=STAGE(I,J)
      CALL MODGRS(M,N,QMAT,M,BMAT,IFLAG) 
      IF (IFLAG.NE.0) THEN
         IFLAG=10
         RETURN
      ENDIF
      CALL YDOT(GETAV,M,N,T+CRK(2)*H,QMAT,STAGE,INARR,REARR)
      CALL DGQTB(M,M,N,QMAT,STAGE,M,BMAT,M)
      DO 12 I=1,N
 12      RKNU2(I)=BMAT(I,I)
 14   CONTINUE
      DO 20 J=1,N
         DO 20 I=1,M
            STAGE(I,J) = Y(I,J) + H*(ARK(3,1)*RK1(I,J)+
     1                 ARK(3,2)*RK2(I,J))
 20   CONTINUE
C ... RK3 AND U45 ...
      CALL YDOT(GETAV,M,N,T+CRK(3)*H,STAGE,RK3,INARR,REARR)
      IF (IHOWLE.EQ.1) GOTO 25
      DO 22 J=1,N
         DO 22 I=1,M
 22         QMAT(I,J)=STAGE(I,J)
      CALL MODGRS(M,N,QMAT,M,BMAT,IFLAG) 
      IF (IFLAG.NE.0) THEN
         IFLAG=10
         RETURN
      ENDIF
      CALL YDOT(GETAV,M,N,T+CRK(3)*H,QMAT,STAGE,INARR,REARR)
      CALL DGQTB(M,M,N,QMAT,STAGE,M,BMAT,M)
      DO 23 I=1,N
 23      RKNU3(I)=BMAT(I,I)
 25   CONTINUE
      DO 30 J=1,N
         DO 30 I=1,M
            STAGE(I,J) = Y(I,J) + H*(ARK(4,1)*RK1(I,J)+
     1                 ARK(4,2)*RK2(I,J)+ARK(4,3)*RK3(I,J))
 30   CONTINUE
C ... RK4 AND U89 
      CALL YDOT(GETAV,M,N,T+CRK(4)*H,STAGE,RK4,INARR,REARR)
      IF (IHOWLE.EQ.1) GOTO 38
      DO 34 J=1,N
         DO 34 I=1,M
 34         QMAT(I,J)=STAGE(I,J)
      CALL MODGRS(M,N,QMAT,M,BMAT,IFLAG) 
      IF (IFLAG.NE.0) THEN
         IFLAG=10
         RETURN
      ENDIF
      CALL YDOT(GETAV,M,N,T+CRK(4)*H,QMAT,STAGE,INARR,REARR)
      CALL DGQTB(M,M,N,QMAT,STAGE,M,BMAT,M)
      DO 35 I=1,N
 35      RKNU4(I)=BMAT(I,I)
 38   CONTINUE
      DO 40 J=1,N
         DO 40 I=1,M
            STAGE(I,J) = Y(I,J) + H*(ARK(5,1)*RK1(I,J)+
     1                 ARK(5,2)*RK2(I,J)+
     2                 ARK(5,3)*RK3(I,J)+ARK(5,4)*RK4(I,J))
 40   CONTINUE
C ... RK5 AND U1 (STAGE VALUE AT H) ...
      CALL YDOT(GETAV,M,N,T+CRK(5)*H,STAGE,RK5,INARR,REARR)
      IF (IHOWLE.EQ.1) GOTO 48
      DO 43 J=1,N
         DO 43 I=1,M
 43         QMAT(I,J)=STAGE(I,J)
      CALL MODGRS(M,N,QMAT,M,BMAT,IFLAG) 
      IF (IFLAG.NE.0) THEN
         IFLAG=10
         RETURN
      ENDIF
      CALL YDOT(GETAV,M,N,T+CRK(5)*H,QMAT,STAGE,INARR,REARR)
      CALL DGQTB(M,M,N,QMAT,STAGE,M,BMAT,M)
      DO 47 I=1,N
 47      RKNU5(I)=BMAT(I,I)
 48   CONTINUE
      DO 50 J=1,N
         DO 50 I=1,M
            STAGE(I,J) = Y(I,J) + H*(ARK(6,1)*RK1(I,J)+
     1                ARK(6,2)*RK2(I,J)+ARK(6,3)*RK3(I,J)+
     2                ARK(6,4)*RK4(I,J)+ARK(6,5)*RK5(I,J))
 50   CONTINUE
C ... RK6, SOLUTION AT H ...
      CALL YDOT(GETAV,M,N,T+H,STAGE,RK6,INARR,REARR)
      IF (IHOWLE.EQ.1) GOTO 56
      DO 52 J=1,N
         DO 52 I=1,M
 52         QMAT(I,J)=STAGE(I,J)
      CALL MODGRS(M,N,QMAT,M,BMAT,IFLAG) 
      IF (IFLAG.NE.0) THEN
         IFLAG=10
         RETURN
      ENDIF
      CALL YDOT(GETAV,M,N,T+H,QMAT,STAGE,INARR,REARR)
      CALL DGQTB(M,M,N,QMAT,STAGE,M,BMAT,M)
      DO 55 I=1,N
 55      RKNU6(I)=BMAT(I,I)
 56   CONTINUE
      DO 61 J=1,N
         IF (IHOWLE.EQ.0) THEN 
            RNU(J)=RNU(J)+H*(ARK(7,1)*RKNU1(J)+
     1         ARK(7,3)*RKNU3(J)+ARK(7,4)*RKNU4(J)+
     2         ARK(7,5)*RKNU5(J)+ARK(7,6)*RKNU6(J))
         ENDIF
         DO 59 I=1,M
            Y(I,J) = Y(I,J) + H*(ARK(7,1)*RK1(I,J)+
     1         ARK(7,3)*RK3(I,J)+ARK(7,4)*RK4(I,J)+
     2         ARK(7,5)*RK5(I,J)+ARK(7,6)*RK6(I,J))
 59      CONTINUE
 61   CONTINUE 
C ... RK7 (STORED IN RK2), AND COMPARISON SOLUTION YHAT AND RNUHAT ...
C TO BE USED FOR ERROR MONITOR/STEP-SIZE SELECTION
      CALL YDOT(GETAV,M,N,T+H,Y,RK2,INARR,REARR)
      CALL MODGRS(M,N,Y,M,BMAT,IFLAG) 
      IF (IFLAG.NE.0) THEN
         IFLAG=10
         RETURN
      ENDIF
      CALL YDOT(GETAV,M,N,T+H,Y,STAGE,INARR,REARR)
      CALL DGQTB(M,M,N,Y,STAGE,M,BMAT,M)
      DO 64 I=1,N
 64      RKNU2(I)=BMAT(I,I)
      DO 73 J=1,N 
         IF (IHOWLE.EQ.0.AND.IFXDPF.EQ.0.AND.IERRC.NE.1) THEN
            RNUHAT(J)=RNUHAT(J)+H*(BHAT(1)*RKNU1(J)+
     1       BHAT(3)*RKNU3(J)+BHAT(4)*RKNU4(J)+
     2       BHAT(5)*RKNU5(J)+BHAT(6)*RKNU6(J)+BHAT(7)*RKNU2(J))
            ELSEIF (IHOWLE.EQ.1) THEN
             CALL TRAPRL(RKNU1(J),RKNU2(J),H,RNU(J))
         ENDIF
         IF (((IERRC.EQ.1).OR.(IERRC.EQ.10)).AND.IFXDPF.EQ.0) THEN
            DO 70 I=1,M
               YHAT(I,J)=YHAT(I,J)+H*(BHAT(1)*RK1(I,J)+
     1          BHAT(3)*RK3(I,J)+BHAT(4)*RK4(I,J)+
     2          BHAT(5)*RK5(I,J)+BHAT(6)*RK6(I,J)+BHAT(7)*RK2(I,J))
 70         CONTINUE
         ENDIF
 73   CONTINUE
      IF (((IERRC.EQ.1).OR.(IERRC.EQ.10)).AND.IFXDPF.EQ.0) THEN
         CALL MODGRS(M,N,YHAT,M,BMAT,IFLAG) 
         IF (IFLAG.NE.0) THEN
            IFLAG=10
            RETURN
         ENDIF
      ENDIF

      RETURN
      END
CCCCC END OF HQRKDP CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: PQ38LL
C
C*****Purpose
C
C performs the one step integration of Q'=F(A,Q), Q0 given
C and monitors errors on the one step approximations to the LEs
C updates stepsize and approximation of the LEs
C
C       Method used: Projection continuous QR method with RK38 as 
C          integrator for Q, plus complete projection for Q (i.e., also
C          for stage values). As specified by
C          IPAR(9), two options are provided for 
C          approximating the LEs:
C          (i) a quadrature rule for the nu-variables of the 
C              same order as RK38 (the default), or
C          (ii) the composite trap-rule.  
C  
C          Code can proceed in fixed or variable
C          step size.  In the latter case, error control is 
C          performed on Q or on the Lyapunov exponents,
C          depending on IPAR(9),
C          No approximation of solution trajectory is attempted.
C 
C*****Author: L Dieci 
C*****Date: May 29, 2004
C
CCCCCC PQ38LL CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE PQ38LL(GETAV,M,N,T,H,HOLD,Y0,RLYAP,TOLQ,TOLL,
     *          Q,QHAT,BMAT,RK1,RK2,RK3,RK4,STAGE, 
     *          RNU,RNUHAT,RKNU1,RKNU2,RKNU3,RKNU4,
     *          ARK,CRK,BHAT,IPAR,NREJ,IFLAG,INARR,REARR)

      IMPLICIT NONE

      EXTERNAL GETAV
      INTEGER M,N,IPAR(*),NREJ,IFLAG, INARR(*)
      DOUBLE PRECISION T,H,HOLD, REARR(*)
      DOUBLE PRECISION Y0(M,N), RLYAP(N), TOLQ, TOLL(*)
      DOUBLE PRECISION Q(M,N), QHAT(M,N)
      DOUBLE PRECISION BMAT(M,N)
      DOUBLE PRECISION RK1(M,N),RK2(M,N)
      DOUBLE PRECISION RK3(M,N),RK4(M,N)
      DOUBLE PRECISION STAGE(M,N)
      DOUBLE PRECISION RNU(N), RNUHAT(N)       
      DOUBLE PRECISION RKNU1(N), RKNU2(N), RKNU3(N), RKNU4(N)       
      DOUBLE PRECISION ARK(7,7), CRK(7), BHAT(7)

      LOGICAL PREVIO, LAST
      INTEGER INTDIR, IFIRST, IFXDPF, IHOWLE, IERRC, I, J
      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      DOUBLE PRECISION HNEW, HMIN, HNEXT, TFIRST
      DOUBLE PRECISION SCALE, RTOL, ERR
      COMMON /STEP/ HMIN, HNEXT, TFIRST, INTDIR, IFIRST, LAST
      SAVE /STEP/
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/

      IFXDPF=IPAR(1)
      IHOWLE=IPAR(9)
      IERRC=IPAR(10)
      PREVIO=.FALSE.
C PREVIO IS .FALSE. IF PREVIOUS STEP WAS SUCCESSFUL
C IT IS .TRUE. IF PREVIOUS STEP WAS A FAILURE
 2    CONTINUE
CCCCCCCCCC INTEGRATE FOR Q: RETURNS Q AND QHAT
      CALL PQRK38(GETAV,PREVIO,M,N,IFIRST,IFXDPF,IHOWLE,IERRC,
     *        T,H,Y0,Q,QHAT,BMAT,RK1,RK2,RK3,RK4,STAGE,
     *        RNU,RNUHAT,RKNU1,RKNU2,RKNU3,RKNU4,
     *        ARK,CRK,BHAT,IFLAG,INARR,REARR)
      IF (IFIRST.EQ.1) IFIRST=0
      IF (IFLAG.NE.0) THEN
         IF (IFXDPF.EQ.0) THEN
C TRY DRASTIC WAY TO RECOVER
            IF (DABS(H).GT.HMIN) THEN
               IFLAG=0
               H=INTDIR*HMIN
               PREVIO=.TRUE.
               NREJ=NREJ+1
               GOTO 2
              ELSE
               IFLAG=5
               RETURN
            ENDIF
           ELSE
            IFLAG=10
            RETURN
         END IF
      END IF
      IF (IFXDPF.EQ.0) THEN
C HERE WE PREDICT NEXT STEPSIZE HNEW, NEVER ALLOWING IT TO
C BE GREATER THAN FIVE TIMES CURRENT H 
C (THIS VALUE CAN BE CHANGED IN HINCR IN "INIT")
         HNEW=HINCR*H
         RTOL=ZERO
         IF (IERRC.EQ.0) THEN
            CALL NERRNU(N,RNU,RNUHAT,ERR,TOLL,ONE,ZERO)
           ELSE IF (IERRC.EQ.1) THEN
            CALL NERRQ(M,N,Q,QHAT,Y0,ERR,TOLQ,ONE,ZERO)
           ELSE 
            CALL NERRNU(N,RNU,RNUHAT,ERR,TOLL,ONE,ZERO)
            RTOL=DMAX1(RTOL,ERR)
            CALL NERRQ(M,N,Q,QHAT,Y0,ERR,TOLQ,ONE,ZERO)
            ERR=DMAX1(ERR,RTOL)
         END IF
C PREDICT NEW STEPSIZE
         IF (ERR.GT.ZERO)         
     1      HNEW=DMIN1(DABS(HNEW),DABS(HSFTY*H*(ONE/ERR)**EXP2))
C GO BACK IF FAILED UNLESS H TOO SMALL
         IF (ERR.GT.ONE) THEN
C AVOID DRASTIC REDUCTION OF STEPSIZE (NEVER LESS THAN 1/5)
C (THIS VALUE CAN BE CHANGED IN HDRSTC IN "INIT")
            H=INTDIR*DMAX1(DABS(HNEW),HDRSTC*DABS(H))
            NREJ=NREJ+1
            IF (INTDIR*H.LE.HMIN) THEN
               PRINT*,' H = ', H
               IFLAG = 5
               RETURN
            ENDIF
            PREVIO=.TRUE.
            GOTO 2
         ENDIF            
      END IF
C HAVE COMPLETED THE STEP WITH STEPSIZE H
C SAVE LAST SUCCESSFUL STEPSIZE
      HOLD=H
      IF (IFXDPF.EQ.0) THEN
C         HNEW=INTDIR*DMAX1(DABS(HNEW),DABS(H))
         IF((PREVIO).AND.(DABS(HNEW).GT.DABS(H)))HNEW=H
         H=HNEW
      END IF
C FINALLY, UPDATE THE APPROXIMATION FOR THE LES AND Q
      SCALE=ONE/(T+HOLD-TFIRST)      
      DO 25 I=1,N
         RLYAP(I)=((T-TFIRST)*SCALE)*RLYAP(I)+SCALE*RNU(I)
C NEXT LINE IS THE HAT-APPROXIMATION (LOWER ORDER)
C         RLYAP(I)=((T-TFIRST)*SCALE)*RLYAP(I)+SCALE*RNUHAT(I)
 25   CONTINUE         
      DO 28 J=1,N
         DO 28 I=1,M
            Y0(I,J)=Q(I,J)
 28   CONTINUE

      RETURN
      END
CCCCC END OF PQ38LL CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: PQRK38
C
C*****Purpose
C
C integrates on one step the solution of Q'=F(A(t),Q), Q0=Q0
C and returns one approximations, Q to the Q-factor
C in the QR factorization of the fundamental solution Y
C It also returns two approximations to the updates for
C the LEs, rnu and rnuhat, which can be used for error purposes
C Integrator is the 3/8 rule with embedded FSAL of order 4 (3)
C Scheme is total projection method if ihowle=0, or the
C the projection of last value only if ihowle=1
C 
C*****Author: L Dieci 
C*****Date: May 29, 2004
C
CCCCCC PQRK38 CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE PQRK38(GETAV,PREVIO,M,N,IFIRST,IFXDPF,IHOWLE,IERRC,
     *        T,H,Y0,Q,QHAT,BMAT,RK1,RK2,RK3,RK4,STAGE,
     *        RNU,RNUHAT,RKNU1,RKNU2,RKNU3,RKNU4,
     *        ARK,CRK,BHAT,IFLAG,INARR,REARR)

      IMPLICIT NONE

      EXTERNAL GETAV
      LOGICAL PREVIO
      INTEGER M, N, IFIRST, IFXDPF, IHOWLE, IERRC, IFLAG, INARR(*)
      DOUBLE PRECISION H, T, Y0(M,N), REARR(*)
      DOUBLE PRECISION Q(M,N), QHAT(M,N)
      DOUBLE PRECISION BMAT(M,N)
      DOUBLE PRECISION RK1(M,N),RK2(M,N),RK3(M,N),RK4(M,N)
      DOUBLE PRECISION STAGE(M,N)
      DOUBLE PRECISION RNU(N),RNUHAT(N)
      DOUBLE PRECISION RKNU1(N),RKNU2(N),RKNU3(N),RKNU4(N)
      DOUBLE PRECISION ARK(7,7), CRK(7), BHAT(7)
      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      INTEGER I, J
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/

C INTIALIZE Q TO OLD Q
C THE RNU'S ARE INITIALLY AT 0, SINCE WE TAKE ONE STEP OF INTEGRAL
      DO 2 J=1,N
         RNU(J)=ZERO
         RNUHAT(J)=ZERO
         DO 2 I=1,M
            Q(I,J)=Y0(I,J)
            QHAT(I,J)=Y0(I,J)
 2    CONTINUE
C FORM RK1, RK2, RK3, RK4 AND RK5 
C FORM ALSO THE STAGE VALUES AT H/3, 2H/3 AND H (NB THE LATTER IS
C NOT THE SOLUTION VALUE, BUT JUST THE STAGE VALUE)
C ... RK1 AND U13 ...
C VERY FIRST CALL WITH FIXED STEPSIZE, OR IF NOT FAILED STEP NEED RK1
C OTHERWISE, HAVE RK1 IN RK1 
CC **** SEE IF CAN ECONOMIZE HERE ****
      CALL QDOT(GETAV,M,N,T,BMAT,Q,RK1,INARR,REARR)
      DO 5 J=1,N
         DO 5 I=1,M
            STAGE(I,J) = Q(I,J) + H*ARK(2,1)*RK1(I,J)
 5    CONTINUE
      DO 10 I=1,N
 10       RKNU1(I)=BMAT(I,I)
      IF (IHOWLE.EQ.0) THEN
         CALL MODGRS(M,N,STAGE,M,BMAT,IFLAG) 
         IF (IFLAG.NE.0) THEN
            IFLAG=10
            RETURN
         ENDIF
      ENDIF
C ... RK2 AND U23 ...
      CALL QDOT(GETAV,M,N,T+CRK(2)*H,BMAT,STAGE,RK2,INARR,REARR)
      DO 12 J=1,N
         DO 12 I=1,M
            STAGE(I,J) = Q(I,J) + H*(ARK(3,1)*RK1(I,J)+
     1                 ARK(3,2)*RK2(I,J))
 12   CONTINUE
      IF (IHOWLE.EQ.0) THEN
         DO 20 I=1,N
 20         RKNU2(I)=BMAT(I,I)
         CALL MODGRS(M,N,STAGE,M,BMAT,IFLAG) 
         IF (IFLAG.NE.0) THEN
            IFLAG=10
            RETURN
         ENDIF
      ENDIF
C ... RK3 AND U1 (STAGE VALUE AT H) ...
      CALL QDOT(GETAV,M,N,T+CRK(3)*H,BMAT,STAGE,RK3,INARR,REARR)
      DO 22 J=1,N
         DO 22 I=1,M
            STAGE(I,J) = Q(I,J) + H*(ARK(4,1)*RK1(I,J)+
     1                ARK(4,2)*RK2(I,J)+ARK(4,3)*RK3(I,J))
 22   CONTINUE
      IF (IHOWLE.EQ.0) THEN
         DO 23 I=1,N
 23         RKNU3(I)=BMAT(I,I)
         CALL MODGRS(M,N,STAGE,M,BMAT,IFLAG) 
         IF (IFLAG.NE.0) THEN
            IFLAG=10
            RETURN
         ENDIF
      ENDIF
C ... RK4, SOLUTION AT H ...
      CALL QDOT(GETAV,M,N,T+H,BMAT,STAGE,RK4,INARR,REARR)
      IF (IHOWLE.EQ.0) THEN
         DO 25 I=1,N
 25         RKNU4(I)=BMAT(I,I)
      ENDIF
      DO 26 J=1,N
         IF (IHOWLE.EQ.0) THEN 
            RNU(J)=RNU(J)+H*(ARK(5,1)*RKNU1(J)+ARK(5,2)*RKNU2(J)+
     1                ARK(5,3)*RKNU3(J)+ARK(5,4)*RKNU4(J))
         ENDIF
         DO 26 I=1,M
            Q(I,J) = Q(I,J) + H*(ARK(5,1)*RK1(I,J)+
     1         ARK(5,2)*RK2(I,J)+ARK(5,3)*RK3(I,J)+ARK(5,4)*RK4(I,J))
 26   CONTINUE
C ... RK5 (STORED IN RK4), AND COMPARISON SOLUTION QHAT AND NUHAT
C TO BE USED FOR ERROR MONITOR/STEP-SIZE SELECTION
      IF (IHOWLE.EQ.0) THEN
         CALL MODGRS(M,N,Q,M,BMAT,IFLAG) 
         IF (IFLAG.NE.0) THEN
            IFLAG=10
            RETURN
         ENDIF
      ENDIF
      IF (((IERRC.EQ.1).OR.(IERRC.EQ.10)).AND.IFXDPF.EQ.0) THEN
         CALL QDOT(GETAV,M,N,T+H,BMAT,Q,RK4,INARR,REARR)
         IF (IHOWLE.EQ.1) THEN
            CALL MODGRS(M,N,Q,M,STAGE,IFLAG) 
            IF (IFLAG.NE.0) THEN
               IFLAG=10
               RETURN
            ENDIF
            CALL YDOT(GETAV,M,N,T+H,Q,STAGE,INARR,REARR)
            CALL DGQTB(M,M,N,Q,STAGE,M,BMAT,M)
         ENDIF
        ELSE
         IF (IHOWLE.EQ.1) THEN
            CALL MODGRS(M,N,Q,M,BMAT,IFLAG) 
            IF (IFLAG.NE.0) THEN
               IFLAG=10
               RETURN
            ENDIF
         ENDIF
         CALL YDOT(GETAV,M,N,T+H,Q,STAGE,INARR,REARR)
         CALL DGQTB(M,M,N,Q,STAGE,M,BMAT,M)
      ENDIF
      DO 28 I=1,N      
 28      RKNU4(I)=BMAT(I,I)
      DO 32 J=1,N 
         IF (IHOWLE.EQ.0.AND.IFXDPF.EQ.0.AND.IERRC.NE.1) THEN
            RNUHAT(J)=RNUHAT(J)+H*(BHAT(1)*RKNU1(J)+BHAT(2)*RKNU2(J)+
     1                BHAT(3)*RKNU3(J)+BHAT(5)*RKNU4(J))
            ELSEIF (IHOWLE.EQ.1) THEN
             CALL TRAPRL(RKNU1(J),RKNU4(J),H,RNU(J))
         ENDIF
         IF (((IERRC.EQ.1).OR.(IERRC.EQ.10)).AND.IFXDPF.EQ.0) THEN
            DO 29 I=1,M
               QHAT(I,J)=QHAT(I,J)+H*(BHAT(1)*RK1(I,J)+
     1           BHAT(2)*RK2(I,J)+BHAT(3)*RK3(I,J)+BHAT(5)*RK4(I,J))
 29         CONTINUE
         ENDIF
 32   CONTINUE
      IF (((IERRC.EQ.1).OR.(IERRC.EQ.10)).AND.IFXDPF.EQ.0) THEN
         CALL MODGRS(M,N,QHAT,M,BMAT,IFLAG) 
         IF (IFLAG.NE.0) THEN
            IFLAG=10
            RETURN
         ENDIF
      ENDIF

      RETURN
      END
CCCCC END OF PQRK38 CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: HQ38LL
C
C*****Purpose
C
C performs the one step integration of Y'=A(t)Y, Y0=Q0
C and monitors errors on the one step approximations to the LEs
C updates stepsize and approximation of the LEs
C 
C       Method used: Hybrid continuous QR method with RK38 as 
C          integrator for Y, extraction of Q.    As specified by
C          IPAR(9), two options are provided for 
C          approximating the LEs:
C          (i) a quadrature rule for the nu-variables of the 
C              same order as RK38 (the default), or
C          (ii) the composite trap-rule.  
C
C          Code can proceed in fixed or variable step size.  
C          In the latter case, error control is 
C          performed on the Lyapunov exponents only or on the Q-factor,
C          depending on IPAR(9).
C          No approximation of solution trajectory is attempted.
C
C*****Author: L Dieci 
C*****Date: May 29, 2004
C
CCCCCC HQ38LL CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE HQ38LL(GETAV,M,N,T,H,HOLD,Y0,RLYAP,TOLQ,TOLL,
     *          Y,YHAT,QMAT,BMAT,RK1,RK2,RK3,RK4,STAGE, 
     *          RNU,RNUHAT,RKNU1,RKNU2,RKNU3,RKNU4,
     *          ARK,CRK,BHAT,IPAR,NREJ,IFLAG,INARR,REARR)

      IMPLICIT NONE

      EXTERNAL GETAV
      INTEGER M,N,IPAR(*),NREJ,IFLAG, INARR(*)
      DOUBLE PRECISION T,H,HOLD, REARR(*)
      DOUBLE PRECISION Y0(M,N), RLYAP(N), TOLQ, TOLL(*)
      DOUBLE PRECISION Y(M,N), YHAT(M,N)
      DOUBLE PRECISION QMAT(M,N), BMAT(M,N)
      DOUBLE PRECISION RK1(M,N),RK2(M,N)
      DOUBLE PRECISION RK3(M,N),RK4(M,N)
      DOUBLE PRECISION STAGE(M,N)
      DOUBLE PRECISION RNU(N), RNUHAT(N)       
      DOUBLE PRECISION RKNU1(N), RKNU2(N), RKNU3(N), RKNU4(N)       
      DOUBLE PRECISION ARK(7,7), CRK(7), BHAT(7)

      LOGICAL PREVIO, LAST
      INTEGER INTDIR, IFIRST, IFXDPF, IHOWLE, IERRC, I, J
      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      DOUBLE PRECISION HNEW, HMIN, HNEXT, TFIRST
      DOUBLE PRECISION SCALE, RTOL, ERR
      COMMON /STEP/ HMIN, HNEXT, TFIRST, INTDIR, IFIRST, LAST
      SAVE /STEP/
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/

      IFXDPF=IPAR(1)
      IHOWLE=IPAR(9)
      IERRC=IPAR(10)
      PREVIO=.FALSE.
C PREVIO IS .FALSE. IF PREVIOUS STEP WAS SUCCESSFUL
C IT IS .TRUE. IF PREVIOUS STEP WAS A FAILURE
 2    CONTINUE
CCCCCCCCCC INTEGRATE FOR Y: RETURNS Q AND QHAT
      CALL HQRK38(GETAV,PREVIO,M,N,IFIRST,IFXDPF,IHOWLE,IERRC,
     *        T,H,Y0,Y,YHAT,QMAT,BMAT,RK1,RK2,RK3,RK4,STAGE,
     *        RNU,RNUHAT,RKNU1,RKNU2,RKNU3,RKNU4,
     *        ARK,CRK,BHAT,IFLAG,INARR,REARR)
      IF (IFIRST.EQ.1) IFIRST=0
      IF (IFLAG.NE.0) THEN
         IF (IFXDPF.EQ.0) THEN
C TRY DRASTIC WAY TO RECOVER
            IF (DABS(H).GT.HMIN) THEN
               IFLAG=0
               H=INTDIR*HMIN
               PREVIO=.TRUE.
               NREJ=NREJ+1
               GOTO 2
              ELSE
               IFLAG=5
               RETURN
            ENDIF
           ELSE
            IFLAG=10
            RETURN
         END IF
      END IF
      IF (IFXDPF.EQ.0) THEN
C HERE WE PREDICT NEXT STEPSIZE HNEW, NEVER ALLOWING IT TO
C BE GREATER THAN FIVE TIMES CURRENT H 
C (THIS VALUE CAN BE CHANGED IN HINCR IN "INIT")
         HNEW=HINCR*H
         RTOL=ZERO
         IF (IERRC.EQ.0) THEN
            CALL NERRNU(N,RNU,RNUHAT,ERR,TOLL,ONE,ZERO)
           ELSE IF (IERRC.EQ.1) THEN
            CALL NERRQ(M,N,Y,YHAT,Y0,ERR,TOLQ,ONE,ZERO)
           ELSE 
            CALL NERRNU(N,RNU,RNUHAT,ERR,TOLL,ONE,ZERO)
            RTOL=DMAX1(RTOL,ERR)
            CALL NERRQ(M,N,Y,YHAT,Y0,ERR,TOLQ,ONE,ZERO)
            ERR=DMAX1(ERR,RTOL)
         END IF
C PREDICT NEW STEPSIZE
         IF (ERR.GT.ZERO)         
     1      HNEW=DMIN1(DABS(HNEW),DABS(HSFTY*H*(ONE/ERR)**EXP2))
C GO BACK IF FAILED UNLESS H TOO SMALL
         IF (ERR.GT.ONE) THEN
C AVOID DRASTIC REDUCTION OF STEPSIZE (NEVER LESS THAN 1/5)
C (THIS VALUE CAN BE CHANGED IN HDRSTC IN "INIT")
            H=INTDIR*DMAX1(DABS(HNEW),HDRSTC*DABS(H))
            NREJ=NREJ+1
            IF (INTDIR*H.LE.HMIN) THEN
               PRINT*,' H = ', H
               IFLAG = 5
               RETURN
            ENDIF
            PREVIO=.TRUE.
            GOTO 2
         ENDIF            
      END IF
C HAVE COMPLETED THE STEP WITH STEPSIZE H
C SAVE LAST SUCCESSFUL STEPSIZE
      HOLD=H
      IF (IFXDPF.EQ.0) THEN
C         HNEW=INTDIR*DMAX1(DABS(HNEW),DABS(H))
         IF((PREVIO).AND.(DABS(HNEW).GT.DABS(H)))HNEW=H
         H=HNEW
      END IF
C FINALLY, UPDATE THE APPROXIMATION FOR THE LES AND Q
      SCALE=ONE/(T+HOLD-TFIRST)      
      DO 25 I=1,N
         RLYAP(I)=((T-TFIRST)*SCALE)*RLYAP(I)+SCALE*RNU(I)
C NEXT LINE IS THE HAT-APPROXIMATION (LOWER ORDER)
C         RLYAP(I)=((T-TFIRST)*SCALE)*RLYAP(I)+SCALE*RNUHAT(I)
 25   CONTINUE         
      DO 28 J=1,N
         DO 28 I=1,M
            Y0(I,J)=Y(I,J)
 28   CONTINUE

      RETURN
      END
CCCCC END OF HQ38LL CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: HQRK38
C
C*****Purpose
C
C integrates on one step the solution of Y'=A(t)Y, Y0=Q0
C and returns two approximations, Y and YHAT
C which can be used for error purposes
C Integrator is the 3/8 rule with embedded FSAL of order 4 (3)
C 
C*****Author: L Dieci 
C*****Date: May 29, 2004
C
CCCCCC HQRK38 CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE HQRK38(GETAV,PREVIO,M,N,IFIRST,IFXDPF,IHOWLE,IERRC,
     *        T,H,Y0,Y,YHAT,QMAT,BMAT,RK1,RK2,RK3,RK4,STAGE,
     *        RNU,RNUHAT,RKNU1,RKNU2,RKNU3,RKNU4,
     *        ARK,CRK,BHAT,IFLAG,INARR,REARR)

      IMPLICIT NONE

      EXTERNAL GETAV
      LOGICAL PREVIO
      INTEGER M, N, IFIRST, IFXDPF, IHOWLE, IERRC, IFLAG, INARR(*)
      DOUBLE PRECISION H, T, Y0(M,N), REARR(*)
      DOUBLE PRECISION Y(M,N), YHAT(M,N)
      DOUBLE PRECISION QMAT(M,N), BMAT(M,N)
      DOUBLE PRECISION RK1(M,N),RK2(M,N),RK3(M,N),RK4(M,N)
      DOUBLE PRECISION STAGE(M,N)
      DOUBLE PRECISION RNU(N),RNUHAT(N)
      DOUBLE PRECISION RKNU1(N),RKNU2(N),RKNU3(N),RKNU4(N)
      DOUBLE PRECISION ARK(7,7), CRK(7), BHAT(7)
      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      INTEGER I, J
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/

C INTIALIZE Y TO OLD Q
C THE RNU'S ARE INITIALLY AT 0, SINCE WE TAKE ONE STEP OF INTEGRAL
      DO 2 J=1,N
         RNU(J)=ZERO
         RNUHAT(J)=ZERO
         DO 2 I=1,M
            Y(I,J)=Y0(I,J)
            YHAT(I,J)=Y0(I,J)
 2    CONTINUE
C FORM RK1, RK2, RK3, RK4 AND RK5 
C FORM ALSO THE STAGE VALUES AT H/3, 2H/3 AND H (NB THE LATTER IS
C NOT THE SOLUTION VALUE, BUT JUST THE STAGE VALUE)
C ... RK1 AND U13 ...
C VERY FIRST CALL WITH FIXED STEPSIZE, OR IF NOT FAILED STEP NEED RK1
C OTHERWISE, HAVE RK1 IN RK1 
      IF ((IFIRST.EQ.1.AND.IFXDPF.EQ.1).OR.(.NOT.PREVIO)) THEN
         CALL YDOT(GETAV,M,N,T,Y0,RK1,INARR,REARR)
      END IF
      CALL DGQTB(M,M,N,Y,RK1,M,BMAT,M)
      DO 4 I=1,N
 4       RKNU1(I)=BMAT(I,I)
      DO 10 J=1,N
         DO 10 I=1,M
            STAGE(I,J) = Y(I,J) + H*ARK(2,1)*RK1(I,J)
 10   CONTINUE
C ... RK2 AND U23 ...
      CALL YDOT(GETAV,M,N,T+CRK(2)*H,STAGE,RK2,INARR,REARR)
      IF (IHOWLE.EQ.1) GOTO 14
      DO 11 J=1,N
         DO 11 I=1,M
 11         QMAT(I,J)=STAGE(I,J)
      CALL MODGRS(M,N,QMAT,M,BMAT,IFLAG) 
      IF (IFLAG.NE.0) THEN
         IFLAG=10
         RETURN
      ENDIF
      CALL YDOT(GETAV,M,N,T+CRK(2)*H,QMAT,STAGE,INARR,REARR)
      CALL DGQTB(M,M,N,QMAT,STAGE,M,BMAT,M)
      DO 12 I=1,N
 12      RKNU2(I)=BMAT(I,I)
 14   CONTINUE
      DO 20 J=1,N
         DO 20 I=1,M
            STAGE(I,J) = Y(I,J) + H*(ARK(3,1)*RK1(I,J)+
     1                 ARK(3,2)*RK2(I,J))
 20   CONTINUE
C ... RK3 AND U1 (STAGE VALUE AT H) ...
      CALL YDOT(GETAV,M,N,T+CRK(3)*H,STAGE,RK3,INARR,REARR)
      IF (IHOWLE.EQ.1) GOTO 25
      DO 21 J=1,N
         DO 21 I=1,M
 21         QMAT(I,J)=STAGE(I,J)
      CALL MODGRS(M,N,QMAT,M,BMAT,IFLAG) 
      IF (IFLAG.NE.0) THEN
         IFLAG=10
         RETURN
      ENDIF
      CALL YDOT(GETAV,M,N,T+CRK(3)*H,QMAT,STAGE,INARR,REARR)
      CALL DGQTB(M,M,N,QMAT,STAGE,M,BMAT,M)
      DO 22 I=1,N
 22      RKNU3(I)=BMAT(I,I)
 25   CONTINUE
      DO 30 J=1,N
         DO 30 I=1,M
            STAGE(I,J) = Y(I,J) + H*(ARK(4,1)*RK1(I,J)+
     1                ARK(4,2)*RK2(I,J)+ARK(4,3)*RK3(I,J))
 30   CONTINUE
C ... RK4, SOLUTION AT H ...
      CALL YDOT(GETAV,M,N,T+H,STAGE,RK4,INARR,REARR)
      IF (IHOWLE.EQ.1) GOTO 38
      DO 34 J=1,N
         DO 34 I=1,M
 34         QMAT(I,J)=STAGE(I,J)
      CALL MODGRS(M,N,QMAT,M,BMAT,IFLAG) 
      IF (IFLAG.NE.0) THEN
         IFLAG=10
         RETURN
      ENDIF
      CALL YDOT(GETAV,M,N,T+H,QMAT,STAGE,INARR,REARR)
      CALL DGQTB(M,M,N,QMAT,STAGE,M,BMAT,M)
      DO 35 I=1,N
 35      RKNU4(I)=BMAT(I,I)
 38   CONTINUE
      DO 40 J=1,N
         IF (IHOWLE.EQ.0) THEN 
            RNU(J)=RNU(J)+H*(ARK(5,1)*RKNU1(J)+ARK(5,2)*RKNU2(J)+
     1                ARK(5,3)*RKNU3(J)+ARK(5,4)*RKNU4(J))
         ENDIF
         DO 40 I=1,M
            Y(I,J) = Y(I,J) + H*(ARK(5,1)*RK1(I,J)+
     1         ARK(5,2)*RK2(I,J)+ARK(5,3)*RK3(I,J)+ARK(5,4)*RK4(I,J))
 40   CONTINUE
C ... RK5 (STORED IN RK4), AND COMPARISON SOLUTION YHAT AND RNUHAT...
C TO BE USED FOR ERROR MONITOR/STEP-SIZE SELECTION
      CALL YDOT(GETAV,M,N,T+H,Y,RK4,INARR,REARR)
      CALL MODGRS(M,N,Y,M,BMAT,IFLAG) 
      IF (IFLAG.NE.0) THEN
         IFLAG=10
         RETURN
      ENDIF
      CALL YDOT(GETAV,M,N,T+H,Y,STAGE,INARR,REARR)
      CALL DGQTB(M,M,N,Y,STAGE,M,BMAT,M)
      DO 41 I=1,N
 41      RKNU4(I)=BMAT(I,I)
      DO 50 J=1,N 
         IF (IHOWLE.EQ.0.AND.IFXDPF.EQ.0.AND.IERRC.NE.1) THEN
            RNUHAT(J)=RNUHAT(J)+H*(BHAT(1)*RKNU1(J)+BHAT(2)*RKNU2(J)+
     1                BHAT(3)*RKNU3(J)+BHAT(5)*RKNU4(J))
            ELSEIF (IHOWLE.EQ.1) THEN
             CALL TRAPRL(RKNU1(J),RKNU4(J),H,RNU(J))
         ENDIF
         IF (((IERRC.EQ.1).OR.(IERRC.EQ.10)).AND.IFXDPF.EQ.0) THEN
            DO 43 I=1,M
               YHAT(I,J)=YHAT(I,J)+H*(BHAT(1)*RK1(I,J)+
     1          BHAT(2)*RK2(I,J)+BHAT(3)*RK3(I,J)+BHAT(5)*RK4(I,J))
 43         CONTINUE
         ENDIF
 50   CONTINUE
      IF (((IERRC.EQ.1).OR.(IERRC.EQ.10)).AND.IFXDPF.EQ.0) THEN
         CALL MODGRS(M,N,YHAT,M,BMAT,IFLAG) 
         IF (IFLAG.NE.0) THEN
            IFLAG=10
            RETURN
         ENDIF
      ENDIF

      RETURN
      END
CCCCC END OF HQRK38 CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: YDPLL
C
C*****Purpose
C
C performs the one step integration of Y'=A(t)Y, Y0=Q0
C and monitors errors on the one step approximations to the LEs
C updates stepsize and approximation of the LEs
C 
C       Method used: discrete QR method with RKDP5 as 
C          integrator for Y.  Can proceed in fixed or variable
C          step size.  In the latter case, error control is 
C          performed on the Lyapunov exponents only.
C          No approximation of solution trajectory is attempted.
C
C*****Author: L Dieci 
C*****Date: May 29, 2004
C
CCCCCC YDPLL CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE YDPLL(GETAV,M,N,T,H,HOLD,Y0,RLYAP,TOL,
     *          Y,YHAT,RK1,RK2,RK3,RK4,RK5,RK6,STAGE, 
     *          ARK,CRK,BHAT,IFXDPF,NREJ,IFLAG,INARR,REARR)

      EXTERNAL GETAV
      INTEGER M,N,IFXDPF,NREJ,IFLAG,INARR(*)
      DOUBLE PRECISION T,H,HOLD,REARR(*)
      DOUBLE PRECISION Y0(M,N), RLYAP(N), TOL(*)
      DOUBLE PRECISION Y(M,N), YHAT(M,N)
      DOUBLE PRECISION RK1(M,N),RK2(M,N)
      DOUBLE PRECISION RK3(M,N),RK4(M,N),RK5(M,N)
      DOUBLE PRECISION RK6(M,N), STAGE(M,N)
      DOUBLE PRECISION ARK(7,7), CRK(7), BHAT(7)

      LOGICAL PREVIO, LAST
      INTEGER INTDIR, IFIRST, I, J
      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      DOUBLE PRECISION HNEW, HMIN, HNEXT, TFIRST
      DOUBLE PRECISION SCALE, RTOL, ERR
      COMMON /STEP/ HMIN, HNEXT, TFIRST, INTDIR, IFIRST, LAST
      SAVE /STEP/
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/

      PREVIO=.FALSE.
C PREVIO IS .FALSE. IF PREVIOUS STEP WAS SUCCESSFUL
C IT IS .TRUE. IF PREVIOUS STEP WAS A FAILURE
 2    CONTINUE
CCCCCCCCCC INTEGRATE FOR Y: RETURNS Y AND YHAT
      CALL YRKDP(GETAV,PREVIO,M,N,IFIRST,IFXDPF,T,H,Y0,
     *        Y,YHAT,RK1,RK2,RK3,RK4,RK5,RK6, STAGE, 
     *        ARK,CRK,BHAT,INARR,REARR)
      IF (IFIRST.EQ.1) IFIRST=0
C NOW DO QR FACTORIZATIONS, GET 1-STEP APPROXIMATIONS
C OF THE LES, CHECK ERRORS AND UPDATE APPROXIMATIONS
C DO QR OF Y AND PUT 1-STEP APPROXIMATIONS OF LES IN RK2(I,I)
C Q-FACTOR OF Y IS IN Y ITSELF
      CALL MODGRS(M,N,Y,M,RK2,IFLAG) 
C      CALL QLAPQR(M,N,Y,M,RK2,STAGE,RK3,IFLAG) 
      IF (IFLAG.NE.0) GOTO 10
      DO 5 I=1,N
         RK2(I,I)=DLOG(RK2(I,I))
 5    CONTINUE
      IF (IFXDPF.EQ.1) GOTO 10
C DO QR OF YHAT AND PUT APPROXIMATE HAT-LES IN STAGE(I,I)
C Q-FACTOR OF YHAT IS IN YHAT
      CALL MODGRS(M,N,YHAT,M,STAGE,IFLAG) 
C      CALL QLAPQR(M,N,YHAT,M,STAGE,RK3,RK4,IFLAG) 
      IF (IFLAG.NE.0) GOTO 10
      DO 7 I=1,N
         STAGE(I,I)=DLOG(STAGE(I,I))
 7    CONTINUE
 10   IF (IFLAG.NE.0) THEN
         IF (IFXDPF.EQ.0) THEN
C TRY DRASTIC WAY TO RECOVER
            IF (DABS(H).GT.HMIN) THEN
               IFLAG=0
               H=INTDIR*HMIN
               PREVIO=.TRUE.
               NREJ=NREJ+1
               GOTO 2
              ELSE
               IFLAG=5
               RETURN
            ENDIF
           ELSE
            IFLAG=10
            RETURN
         END IF
      END IF
      IF (IFXDPF.EQ.0) THEN
C HERE WE PREDICT NEXT STEPSIZE HNEW, NEVER ALLOWING IT TO
C BE GREATER THAN FIVE TIMES CURRENT H 
C (THIS VALUE CAN BE CHANGED IN HINCR IN "INIT")
         HNEW=HINCR*H
C (A): CHECK ERROR RELATIVE TO THE I-TH LES 
C (B): PREDICT I-TH STEPSIZE RESULTING FROM THIS ERROR ESTIMATE
C (C): IF UNSUCCESSFUL, GO BACK 
         RTOL=ZERO
         DO 20 I=1,N
C (A) CHECK ERROR I-TH LE: MIXED ABSOLUTE/RELATIVE ERROR
C     WEIGHTED AGAINST MAGNITUDE
            SCALE=ONE+DABS(RK2(I,I))
            ERR=DABS(RK2(I,I)-STAGE(I,I))/SCALE
            ERR=ERR/TOL(I)
            RTOL=DMAX1(RTOL,ERR)
C (B) PREDICT I-TH STEPSIZE; USE SAFETY FACTOR OF 0.8
C (THIS VALUE CAN BE CHANGED IN HSFTY IN "INIT")
            IF (ERR.GT.ZERO)         
     1         HNEW=DMIN1(DABS(HNEW),DABS(HSFTY*H*(ONE/ERR)**EXP1))
 20      CONTINUE
C (C) GO BACK IF FAILED UNLESS H TOO SMALL
         IF (RTOL.GT.ONE) THEN
C AVOID DRASTIC REDUCTION OF STEPSIZE (NEVER LESS THAN 1/5)
C (THIS VALUE CAN BE CHANGED IN HDRSTC IN "INIT")
            H=INTDIR*DMAX1(DABS(HNEW),HDRSTC*DABS(H))
            NREJ=NREJ+1
            IF (INTDIR*H.LE.HMIN) THEN
               PRINT*,' H = ', H
               IFLAG = 5
               RETURN
            ENDIF
            PREVIO=.TRUE.
            GOTO 2
         ENDIF            
      END IF
C HAVE COMPLETED THE STEP WITH STEPSIZE H
C SAVE LAST SUCCESSFUL STEPSIZE
      HOLD=H
      IF (IFXDPF.EQ.0) THEN
C         HNEW=INTDIR*DMAX1(DABS(HNEW),DABS(H))
         IF((PREVIO).AND.(DABS(HNEW).GT.DABS(H)))HNEW=H
         H=HNEW
      END IF
C FINALLY, UPDATE THE APPROXIMATION FOR THE LES AND Q
      SCALE=ONE/(T+HOLD-TFIRST)      
      DO 25 I=1,N
         RLYAP(I)=((T-TFIRST)*SCALE)*RLYAP(I)+SCALE*RK2(I,I)
C LINE BELOW IS LOWER ORDER (4-TH) APPROXIMATION
C         RLYAP(I)=((T-TFIRST)*SCALE)*RLYAP(I)+SCALE*STAGE(I,I)
 25   CONTINUE         
      DO 28 J=1,N
         DO 28 I=1,M
            Y0(I,J)=Y(I,J)
 28   CONTINUE

      RETURN
      END
CCCCC END OF YDPLL CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: YRKDP
C
C*****Purpose
C
C integrates on one step the solution of Y'=A(t)Y, Y0=Q0
C and returns two approximations, Y and YHAT
C which can be used for error purposes
C Integrator is the Dormand-Prince pair of order 5 (4)
C 
C*****Author: L Dieci 
C*****Date: May 29, 2004
C
CCCCCC YRKDP CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE YRKDP(GETAV,PREVIO,M,N,IFIRST,IFXDPF,T,H,Y0,Y,
     *        YHAT,RK1,RK2,RK3,RK4,RK5,RK6, STAGE,
     *        ARK,CRK,BHAT,INARR,REARR)

      IMPLICIT NONE

      EXTERNAL GETAV
      LOGICAL PREVIO
      INTEGER M, N, IFIRST, IFXDPF, INARR(*)
      DOUBLE PRECISION H, T, Y0(M,N), REARR(*)
      DOUBLE PRECISION Y(M,N),YHAT(M,N)
      DOUBLE PRECISION RK1(M,N),RK2(M,N)
      DOUBLE PRECISION RK3(M,N),RK4(M,N),RK5(M,N)
      DOUBLE PRECISION RK6(M,N), STAGE(M,N)
      DOUBLE PRECISION ARK(7,7), CRK(7), BHAT(7)
      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      INTEGER I, J
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/

C INTIALIZE Y TO OLD Q
      DO 2 J=1,N
         DO 2 I=1,M
            Y(I,J)=Y0(I,J)
            YHAT(I,J)=Y0(I,J)
 2    CONTINUE
C FORM RK1, RK2, RK3, RK4, RK5, RK6, RK7 
C FORM ALSO THE STAGE VALUES AT H/5, 3H/10, 4H/5, 8H/9 AND H (NB THE LATTER IS
C NOT THE SOLUTION VALUE, BUT JUST THE STAGE VALUE)
C ... RK1 AND U15 ...
C VERY FIRST CALL WITH FIXED STEPSIZE, OR IF NOT FAILED STEP NEED RK1
C OTHERWISE, HAVE RK1 IN RK1 
      IF ((IFIRST.EQ.1.AND.IFXDPF.EQ.1).OR.(.NOT.PREVIO)) THEN
         CALL YDOT(GETAV,M,N,T,Y0,RK1,INARR,REARR)
      END IF
      DO 10 J=1,N
         DO 10 I=1,M
            STAGE(I,J) = Y(I,J) + H*ARK(2,1)*RK1(I,J)
 10   CONTINUE
C ... RK2 AND U03 ...
      CALL YDOT(GETAV,M,N,T+CRK(2)*H,STAGE,RK2,INARR,REARR)
      DO 20 J=1,N
         DO 20 I=1,M
            STAGE(I,J) = Y(I,J) + H*(ARK(3,1)*RK1(I,J)+
     1                 ARK(3,2)*RK2(I,J))
 20   CONTINUE
C ... RK3 AND U45 ...
      CALL YDOT(GETAV,M,N,T+CRK(3)*H,STAGE,RK3,INARR,REARR)
      DO 22 J=1,N
         DO 22 I=1,M
            STAGE(I,J) = Y(I,J) + H*(ARK(4,1)*RK1(I,J)+
     1                 ARK(4,2)*RK2(I,J)+ARK(4,3)*RK3(I,J))
 22   CONTINUE
C ... RK4 AND U89 
      CALL YDOT(GETAV,M,N,T+CRK(4)*H,STAGE,RK4,INARR,REARR)
      DO 23 J=1,N
         DO 23 I=1,M
            STAGE(I,J) = Y(I,J) + H*(ARK(5,1)*RK1(I,J)+
     1                 ARK(5,2)*RK2(I,J)+
     2                 ARK(5,3)*RK3(I,J)+ARK(5,4)*RK4(I,J))
 23   CONTINUE
C ... RK5 AND U1 (STAGE VALUE AT H) ...
      CALL YDOT(GETAV,M,N,T+CRK(5)*H,STAGE,RK5,INARR,REARR)
      DO 24 J=1,N
         DO 24 I=1,M
            STAGE(I,J) = Y(I,J) + H*(ARK(6,1)*RK1(I,J)+
     1                ARK(6,2)*RK2(I,J)+ARK(6,3)*RK3(I,J)+
     2                ARK(6,4)*RK4(I,J)+ARK(6,5)*RK5(I,J))
 24   CONTINUE
C ... RK6, SOLUTION AT H ...
      CALL YDOT(GETAV,M,N,T+H,STAGE,RK6,INARR,REARR)
      DO 25 J=1,N
         DO 25 I=1,M
            Y(I,J) = Y(I,J) + H*(ARK(7,1)*RK1(I,J)+
     1         ARK(7,3)*RK3(I,J)+ARK(7,4)*RK4(I,J)+
     2         ARK(7,5)*RK5(I,J)+ARK(7,6)*RK6(I,J))
 25   CONTINUE
      IF (IFXDPF.EQ.1) RETURN
C ... RK7 (STORED IN RK2), AND COMPARISON SOLUTION YHAT ...
C TO BE USED FOR ERROR MONITOR/STEP-SIZE SELECTION
      CALL YDOT(GETAV,M,N,T+H,Y,RK2,INARR,REARR)
      DO 28 J=1,N 
         DO 28 I=1,M
            YHAT(I,J)=YHAT(I,J)+H*(BHAT(1)*RK1(I,J)+
     1       BHAT(3)*RK3(I,J)+BHAT(4)*RK4(I,J)+
     2       BHAT(5)*RK5(I,J)+BHAT(6)*RK6(I,J)+BHAT(7)*RK2(I,J))
 28   CONTINUE

      RETURN
      END
CCCCC END OF YRKDP CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: Y38LL
C
C*****Purpose
C
C performs the one step integration of Y'=A(t)Y, Y0=Q0
C and monitors errors on the one step approximations to the LEs
C updates stepsize and approximation of the LEs
C
C Method used: discrete QR method with RK38 as 
C     integrator for Y.  Can proceed in fixed or variable
C     step size.  In the latter case, error control is 
C     performed on the Lyapunov exponents only.
C     No approximation of solution trajectory is attempted.
C 
C*****Author: L Dieci 
C*****Date: May 29, 2004
C
CCCCCC Y38LL CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE Y38LL(GETAV,M,N,T,H,HOLD,Y0,RLYAP,TOL,
     *          Y,YHAT,RK1,RK2,RK3,RK4,STAGE, 
     *          ARK,CRK,BHAT,IFXDPF,NREJ,IFLAG,INARR,REARR)

      IMPLICIT NONE

      EXTERNAL GETAV
      INTEGER M,N,IFXDPF,NREJ,IFLAG, INARR(*)
      DOUBLE PRECISION T,H,HOLD, REARR(*)
      DOUBLE PRECISION Y0(M,N), RLYAP(N), TOL(*)
      DOUBLE PRECISION Y(M,N), YHAT(M,N)
      DOUBLE PRECISION RK1(M,N),RK2(M,N)
      DOUBLE PRECISION RK3(M,N),RK4(M,N)
      DOUBLE PRECISION STAGE(M,N)
      DOUBLE PRECISION ARK(7,7), CRK(7), BHAT(7)

      LOGICAL PREVIO, LAST
      INTEGER INTDIR, IFIRST, I, J
      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      DOUBLE PRECISION HNEW, HMIN, HNEXT, TFIRST
      DOUBLE PRECISION SCALE, RTOL, ERR
      COMMON /STEP/ HMIN, HNEXT, TFIRST, INTDIR, IFIRST, LAST
      SAVE /STEP/
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/


      PREVIO=.FALSE.
C PREVIO IS .FALSE. IF PREVIOUS STEP WAS SUCCESSFUL
C IT IS .TRUE. IF PREVIOUS STEP WAS A FAILURE
 2    CONTINUE
CCCCCCCCCC INTEGRATE FOR Y: RETURNS Y AND YHAT
      CALL YRK38(GETAV,PREVIO,M,N,IFIRST,IFXDPF,T,H,Y0,
     *        Y,YHAT,RK1,RK2,RK3,RK4,STAGE,
     *        ARK,CRK,BHAT,INARR,REARR)
      IF (IFIRST.EQ.1) IFIRST=0
C NOW DO QR FACTORIZATIONS, GET 1-STEP APPROXIMATIONS
C OF THE LES, CHECK ERRORS AND UPDATE APPROXIMATIONS
C DO QR OF Y AND PUT 1-STEP APPROXIMATIONS OF LES IN RK2(I,I)
C Q-FACTOR OF Y IS IN Y ITSELF
      CALL MODGRS(M,N,Y,M,RK2,IFLAG) 
C      CALL QLAPQR(M,N,Y,M,RK2,STAGE,RK3,IFLAG) 
      IF (IFLAG.NE.0) GOTO 10
      DO 5 I=1,N
         RK2(I,I)=DLOG(RK2(I,I))
 5    CONTINUE
      IF (IFXDPF.EQ.1) GOTO 10
C DO QR OF YHAT AND PUT APPROXIMATE HAT-LES IN U1(I,I)
C Q-FACTOR OF YHAT IS IN YHAT
      CALL MODGRS(M,N,YHAT,M,STAGE,IFLAG) 
C      CALL QLAPQR(M,N,YHAT,M,STAGE,RK3,RK4,IFLAG) 
      IF (IFLAG.NE.0) GOTO 10
      DO 7 I=1,N
         STAGE(I,I)=DLOG(STAGE(I,I))
 7    CONTINUE
 10   IF (IFLAG.NE.0) THEN
         IF (IFXDPF.EQ.0) THEN
C TRY DRASTIC WAY TO RECOVER
            IF (DABS(H).GT.HMIN) THEN
               IFLAG=0
               H=INTDIR*HMIN
               PREVIO=.TRUE.
               NREJ=NREJ+1
               GOTO 2
              ELSE
               IFLAG=5
               RETURN
            ENDIF
           ELSE
            IFLAG=10
            RETURN
         END IF
      END IF
      IF (IFXDPF.EQ.0) THEN
C HERE WE PREDICT NEXT STEPSIZE HNEW, NEVER ALLOWING IT TO
C BE GREATER THAN FIVE TIMES CURRENT H 
C (THIS VALUE CAN BE CHANGED IN HINCR IN "INIT")
         HNEW=HINCR*H
C (A): CHECK ERROR RELATIVE TO THE I-TH LES 
C (B): PREDICT I-TH STEPSIZE RESULTING FROM THIS ERROR ESTIMATE
C (C): IF UNSUCCESSFUL, GO BACK 
         RTOL=ZERO
         DO 20 I=1,N
C (A) CHECK ERROR I-TH LE: MIXED ABSOLUTE/RELATIVE ERROR
C     WEIGHTED AGAINST MAGNITUDE
            SCALE=ONE+DABS(RK2(I,I))
            ERR=DABS(RK2(I,I)-STAGE(I,I))/SCALE
            ERR=ERR/TOL(I)
            RTOL=DMAX1(RTOL,ERR)
C (B) PREDICT I-TH STEPSIZE; USE SAFETY FACTOR OF 0.8
C (THIS VALUE CAN BE CHANGED IN HSFTY IN "INIT")
            IF (ERR.GT.ZERO)         
     1         HNEW=DMIN1(DABS(HNEW),DABS(HSFTY*H*(ONE/ERR)**EXP2))
 20      CONTINUE
C (C) GO BACK IF FAILED UNLESS H TOO SMALL
         IF (RTOL.GT.ONE) THEN
C AVOID DRASTIC REDUCTION OF STEPSIZE (NEVER LESS THAN 1/5)
C (THIS VALUE CAN BE CHANGED IN HDRSTC IN "INIT")
            H=INTDIR*DMAX1(DABS(HNEW),HDRSTC*DABS(H))
            NREJ=NREJ+1
            IF (INTDIR*H.LE.HMIN) THEN
               PRINT*,' H = ', H
               IFLAG = 5
               RETURN
            ENDIF
            PREVIO=.TRUE.
            GOTO 2
         ENDIF            
      END IF
C HAVE COMPLETED THE STEP WITH STEPSIZE H
C SAVE LAST SUCCESSFUL STEPSIZE
      HOLD=H
      IF (IFXDPF.EQ.0) THEN
C         HNEW=INTDIR*DMAX1(DABS(HNEW),DABS(H))
         IF((PREVIO).AND.(DABS(HNEW).GT.DABS(H)))HNEW=H
         H=HNEW
      END IF
C FINALLY, UPDATE THE APPROXIMATION FOR THE LES AND Q
      SCALE=ONE/(T+HOLD-TFIRST)      
      DO 25 I=1,N
         RLYAP(I)=((T-TFIRST)*SCALE)*RLYAP(I)+SCALE*RK2(I,I)
C NEXT LINE IS THE HAT-APPROXIMATION (LOWER ORDER)
C         RLYAP(I)=((T-TFIRST)*SCALE)*RLYAP(I)+SCALE*STAGE(I,I)
 25   CONTINUE         
      DO 28 J=1,N
         DO 28 I=1,M
            Y0(I,J)=Y(I,J)
 28   CONTINUE

      RETURN
      END
CCCCC END OF Y38LL CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: YRK38
C
C*****Purpose
C
C integrates on one step the solution of Y'=A(t)Y, Y0=Q0
C and returns two approximations, Y and YHAT
C which can be used for error purposes
C Integrator is the 3/8 rule with embedded FSAL of order 4 (3)
C 
C*****Author: L Dieci 
C*****Date: May 29, 2004
C
CCCCCC YRK38 CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE YRK38(GETAV,PREVIO,M,N,IFIRST,IFXDPF,T,H,Y0,Y,
     *        YHAT,RK1,RK2,RK3,RK4,STAGE,ARK,CRK,BHAT,INARR,REARR)

      IMPLICIT NONE

      EXTERNAL GETAV
      LOGICAL PREVIO
      INTEGER M, N, IFIRST, IFXDPF, INARR(*)
      DOUBLE PRECISION H, T, Y0(M,N), REARR(*)
      DOUBLE PRECISION Y(M,N), YHAT(M,N)
      DOUBLE PRECISION RK1(M,N),RK2(M,N),RK3(M,N),RK4(M,N)
      DOUBLE PRECISION STAGE(M,N)
      DOUBLE PRECISION ARK(7,7), CRK(7), BHAT(7)
      DOUBLE PRECISION ONE, ZERO, EXP1,EXP2
      DOUBLE PRECISION HINCR, HSFTY, HDRSTC,TEPS
      INTEGER I, J
      COMMON /CONSTS/ONE,ZERO,EXP1,EXP2,HINCR,HSFTY,HDRSTC,TEPS
      SAVE /CONSTS/

C INTIALIZE Y TO OLD Q
      DO 2 J=1,N
         DO 2 I=1,M
            Y(I,J)=Y0(I,J)
            YHAT(I,J)=Y0(I,J)
 2    CONTINUE
C FORM RK1, RK2, RK3, RK4 AND RK5 
C FORM ALSO THE STAGE VALUES AT H/3, 2H/3 AND H (NB THE LATTER IS
C NOT THE SOLUTION VALUE, BUT JUST THE STAGE VALUE)
C ... RK1 AND U13 ...
C VERY FIRST CALL WITH FIXED STEPSIZE, OR IF NOT FAILED STEP NEED RK1
C OTHERWISE, HAVE RK1 IN RK1 
      IF ((IFIRST.EQ.1.AND.IFXDPF.EQ.1).OR.(.NOT.PREVIO)) THEN
         CALL YDOT(GETAV,M,N,T,Y0,RK1,INARR,REARR)
      END IF
      DO 10 J=1,N
         DO 10 I=1,M
            STAGE(I,J) = Y(I,J) + H*ARK(2,1)*RK1(I,J)
 10   CONTINUE
C ... RK2 AND U23 ...
      CALL YDOT(GETAV,M,N,T+CRK(2)*H,STAGE,RK2,INARR,REARR)
      DO 20 J=1,N
         DO 20 I=1,M
            STAGE(I,J) = Y(I,J) + H*(ARK(3,1)*RK1(I,J)+
     1                 ARK(3,2)*RK2(I,J))
 20   CONTINUE
C ... RK3 AND U1 (STAGE VALUE AT H) ...
      CALL YDOT(GETAV,M,N,T+CRK(3)*H,STAGE,RK3,INARR,REARR)
      DO 30 J=1,N
         DO 30 I=1,M
            STAGE(I,J) = Y(I,J) + H*(ARK(4,1)*RK1(I,J)+
     1                ARK(4,2)*RK2(I,J)+ARK(4,3)*RK3(I,J))
 30   CONTINUE
C ... RK4, SOLUTION AT H ...
      CALL YDOT(GETAV,M,N,T+H,STAGE,RK4,INARR,REARR)
      DO 40 J=1,N
         DO 40 I=1,M
            Y(I,J) = Y(I,J) + H*(ARK(5,1)*RK1(I,J)+
     1         ARK(5,2)*RK2(I,J)+ARK(5,3)*RK3(I,J)+ARK(5,4)*RK4(I,J))
 40   CONTINUE
      IF (IFXDPF.EQ.1) RETURN
C ... RK5 (STORED IN RK4), AND COMPARISON SOLUTION YHAT ...
C TO BE USED FOR ERROR MONITOR/STEP-SIZE SELECTION
      CALL YDOT(GETAV,M,N,T+H,Y,RK4,INARR,REARR)
      DO 50 J=1,N 
         DO 50 I=1,M
            YHAT(I,J)=YHAT(I,J)+H*(BHAT(1)*RK1(I,J)+
     1       BHAT(2)*RK2(I,J)+BHAT(3)*RK3(I,J)+BHAT(5)*RK4(I,J))
 50   CONTINUE

      RETURN
      END
CCCCC END OF YRK38 CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: NERRQ
C
C*****Purpose
C     Computes max-element norm of errors of Y by columns
C     using a mixed absolute/relative error criterion
C
C*****Authors: L Dieci 
C*****Date: April 23, 2004
C
CCCCCCC NERRQ CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE NERRQ(M,N,Q,QHAT,Y0,ERR,TOL,ONE,ZERO)

      IMPLICIT NONE

      INTEGER M,N
      DOUBLE PRECISION Q(M,N), QHAT(M,N), Y0(M,N), TOL, ERR
      DOUBLE PRECISION ONE, ZERO

      DOUBLE PRECISION RTOL, SCALE
      INTEGER I,J

C Computes the weighted norm of the error for step-size selection
C Uses the 2-norm for each column and then the max of these errors

      RTOL=ZERO
      DO 20 J=1,N
         ERR=ZERO
C WEIGH AGAINST LARGEST OF INITIAL AND FINAL VALUES
         DO 10 I=1,M
            SCALE=ONE+DMAX1(DABS(Q(I,J)),DABS(Y0(I,J)))
            ERR=DMAX1(ERR,DABS(Q(I,J)-QHAT(I,J))/SCALE)
 10      CONTINUE
         ERR=ERR/TOL
         RTOL=DMAX1(RTOL,ERR)
 20   CONTINUE
      ERR=RTOL

      RETURN 
      END
CCCCC END OF NERRQ CCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C*****Subroutine name: NERRNU
C
C*****Purpose
C     Computes supnorm of errors of nu (integrals)
C     using a mixed absolute/relative error criterion
C
C*****Authors: L Dieci 
C*****Date: April 23, 2004
C
CCCCCCC NERRNU CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE NERRNU(N,RNU,RNUHAT,ERR,TOL,ONE,ZERO)

      IMPLICIT NONE

      INTEGER N
      DOUBLE PRECISION RNU(N), RNUHAT(N), TOL(*), ERR
      DOUBLE PRECISION ONE, ZERO

      DOUBLE PRECISION RTOL, SCALE
      INTEGER I

C Computes the weighted sup-norm of the error for step-size selection

      RTOL=ZERO
      DO 10 I=1,N
C RECALL THAT INITIAL VALUE IS 0
         SCALE=ONE+DABS(RNU(I))
         ERR=DABS(RNU(I)-RNUHAT(I))/SCALE
         ERR=ERR/TOL(I)
         RTOL=DMAX1(RTOL,ERR)
 10   CONTINUE
      ERR=RTOL

      RETURN 
      END
CCCCC END OF NERRNU CCCCC
